/* ==== ai_waypt.c ==== */
/* To be merged into ai.c */
/* Or not. */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




#include <stdlib.h>
#include "C-foogrid.h"
#include "porrog_defs.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h"
#include "ai_t.h"
#include "ai_wpt.h"

/* protos */
#include "ai_p.h"
#include "mg_utils.h"
#include "memutils.h"
#include "ui.h"
#include "crash.h"

/* Ways to do things:
 Now thinking that U shaped routes can't be figured out by
 the "find grid elements shared by mob's tunnels and target's tunnels"
 method. So instead, I'm thinking the method could just be used as
 a heuristic to order a search?
 Alternatively, it might not make a search faster:
 I was never convinced the grid method *was* generally
 faster than a search, merely that it had bounded complexity
 (which is clearly true).
 Search is painful currently. It could be a LOT less painful
 if I introduced "intersection" structures that could be introduced
 with regions, to find the exact points where intersections occur
 in tunnels: Waypoints in tunnels ought to be intersections,
 and figuring out which coordinates *are* intersections is relatively
 intensive :(*/

tunnelsearch_t ts_array[20]; /* This array must have .Freeit elements set 0 */
intersearch_t is_array[20]; /* Ditto. */
int tscount=0;
int iscount=0;
/* Macros for mapping coords of tunnel sections to 2 bit grid coords */
#define XCOORDS_TO_XELEM(tx) ( (tx>=gsx2)?((tx>=gsx3)?3:2):((tx>=gsx1)?1:0) )
#define YCOORDS_TO_YELEM(ty) ( (ty>=gsy2)?((ty>=gsy3)?3:2):((ty>=gsy1)?1:0) )
/* Not 100% clear if I should be using >= or > for comparison. I think
that >= is more suitable. */

/* Sets a waypoint for a mob in a location of type mstype to reach (tx,ty),
which should (we hope) be in the same location...
if d is set, target is a door. Else could be player, or another mob?? */
void ai_set_waypoint(monster_t *mob,int mstype,int tx,int ty,int d)
{
    int mx=mob->X,my=mob->Y;
    int p,blocked=0;
    region_t *msregion;
    int tx1,tx2,ty1,ty2; /* Temp vars for tunnel coords */
    int ix,iy; /* Temp vars of intersection */
    int mt_tot=0,tt_tot=0; /* Total entries in following */
    tunnellist_t *mtunnels[4]; /* Tunnels mob stands in */
    tunnellist_t *ttunnels[4]; /* Tunnels target is in */
    tunnellist_t *tun;
    intersection_t *inter;
    intersection_t *wpt;

    intersearch_t *icheck=NULL; /*intersections to check */
    intersearch_t *ichecked=NULL; /*intersections already checked */
    tunnelsearch_t *tcheck=NULL; /* tunnels to check */
    tunnelsearch_t *tchecked=NULL; /*tunnels already checked */
    intersearch_t *isearch;
    tunnelsearch_t *tsearch;

    maploc_t *mapstart=map->Data;
    tile_t tile;
/*  if (mstype==2) { Should be either 0 or 1 } */

    switch (mstype) {
       case 0: /* Tunnel regions- the most complicated! */
         /* Of intersections between mob and target, find one
         closest to mob. If mob has LOS to target, set waypoint
         to target. If can't work out intersections between mob
         and target, pick some random intersection in LOS of
         mob, prolly other than the closest to/under mob
         (which should be the one the mob is coming from) */
         if (mx==tx) {
            /* Vertically aligned */
            if (my<ty) {
               for(p=my;p<ty;p++){
                 tile=(mapstart+FOOGRID_IDX(map,p,mx))->Tiletype;
                 if(tile==PTL_PWALL||tile==PTL_ROCK||tile==PTL_HWALL||tile==PTL_VWALL||tile==PTL_CORNER)blocked=1;
               }
            } else {
               for(p=ty;p<my;p++){
                 tile=(mapstart+FOOGRID_IDX(map,p,mx))->Tiletype;
                 if(tile==PTL_PWALL||tile==PTL_ROCK||tile==PTL_HWALL||tile==PTL_VWALL||tile==PTL_CORNER)blocked=1;
               }
            }
            /* How do we say what to do if a blockage were found? */
            if (blocked==0) {mob->Wx=tx;mob->Wy=ty;return;}
            /* Easy! We fall through! */
         } else if (my==ty) {
            /* Horizontally aligned */
            if(mx<tx) {
               for(p=mx;p<tx;p++) {
                 tile=(mapstart+FOOGRID_IDX(map,my,p))->Tiletype;
                 if(tile==PTL_PWALL||tile==PTL_ROCK||tile==PTL_HWALL||tile==PTL_VWALL||tile==PTL_CORNER)blocked=1;
               }
            } else {
               for(p=tx;p<mx;p++) {
                 tile=(mapstart+FOOGRID_IDX(map,my,p))->Tiletype;
                 if(tile==PTL_PWALL||tile==PTL_ROCK||tile==PTL_HWALL||tile==PTL_VWALL||tile==PTL_CORNER)blocked=1;
               }
            }
            if (blocked==0) {mob->Wx=tx;mob->Wy=ty;return;}
         }
         /* Need to find some complex path :P */

         msregion=find_region(mx,my);
/*       if (d) { It's a door. Why do we care again?
         } */

         /* Put all tunnels that cross (mx,my) into a set MT.
         Put all tunnels that cross (tx,ty) into a set TT. THEN:
         Also, tunnels in MT should go into CHECKED list.*/

         /* OLD METHOD: Bounded time, but slow and not guaranteed to work.
         In the end, I couldn't justify its use really.*/

         /* NEW method: Unbounded time, but prolly not
         that slow unless region is gigantic.
         Unlikely to be more than around 2 iterations.

         Find intersections with all of MT, put into a list
         PWP (potential waypoints). Exclude any that are at
         mx,my. NB: PWP would actually not be a specific
         list, but would juse be the list of intersections
         to check, with the "waypoint" pointer pointing to
         the same intersection as "This"!
         Now, for each intersection to check, find all tunnels
         that those points are in, tag those tunnels with the
         same waypoint as the intersection they came from.
         Ignore any that are in the "checked list" already,
         and the rest should go into tunnel-check list.
         Oh, before that, check if any are in TT-
         because if so, that's a path, and we set the
         associated intersection point to be the monster's
         waypoint.
         If none were in TT, just use the tunnels in CHECK
         to produce new intersection points (again, tagged
         with the identifiers of the original intersecion
         from PWP). Use these to generate new tunnels
         to go into CHECK. */
         for(p=0;p<4;p++){mtunnels[p]=NULL;ttunnels[p]=NULL;}

         for(tun=msregion->Firsttunnel;tun!=NULL;tun=tun->Nexttunnel)
         {
            tx1=tun->Xstart;tx2=tun->Xend;
            ty1=tun->Ystart;ty2=tun->Yend;
            if( (tx1<=mx)&&(mx<=tx2)
            &&  (ty1<=my)&&(my<=ty2)) {
               mtunnels[mt_tot]=tun;mt_tot++;
               tsearch=new_tunnelsearch();
               tsearch->Next=tcheck;
               tcheck=tsearch;
               tsearch->This=tun;
               /* Can we set the Waypoint element? */
            }
            if( (tx1<=tx)&&(tx<=tx2)
            &&  (ty1<=ty)&&(ty<=ty2)) {
               ttunnels[tt_tot]=tun;tt_tot++;
            }
         }
         /* Should change the above to use the get_tunnel() fn*/
         /* Scan through all of MT (now in tcheck) to find any
         intersections not at mx,my. Put those in icheck, with
         Wapoint=This. */
         for(;tcheck!=NULL;
         tsearch=tcheck,tcheck=tcheck->Next,
         tsearch->Next=tchecked,tchecked=tsearch/*mv tcheck to tchecked*/)
         {
            tun=tcheck->This;
            for(inter=get_intersections(tun,msregion->Intersections);
             inter!=NULL;
             inter=get_intersections(tun,inter->Next)){
               if (inter->Xcoord==mx && inter->Ycoord==my) continue;
               isearch=new_intersearch();
               isearch->This=inter;isearch->Waypoint=inter;
               isearch->Next=icheck;
               icheck=isearch;
            }
         }
         /* Now, start a loop until we hit the target
         or run out of intersections to search. */
         while(icheck!=NULL) {
           for(;icheck!=NULL;
           isearch=icheck,icheck=icheck->Next,
           isearch->Next=ichecked,ichecked=isearch/*mv icheck to ichecked*/)
           {
             inter=icheck->This;wpt=icheck->Waypoint;
             ix=inter->Xcoord;iy=inter->Ycoord;
             for(tun=get_tunnels(msregion->Firsttunnel,ix,iy);
              tun!=NULL;
              tun=get_tunnels(tun->Nexttunnel,ix,iy)){
                /* Find if tun is in tchecked, if so skip to next tun*/
                p=0;
                for(tsearch=tchecked;tsearch!=NULL;tsearch=tsearch->Next)
                   if(tsearch->This==tun) {p=1;continue;}
                if(p==1/*in_tchecked(tun)*/) continue;
                /* If tun is in TT, we've got a path, so set waypoint to wpt*/
                for(p=0;p<tt_tot;p++) if(ttunnels[p]==tun) {p=-1;break;}
                if(p==-1/*in_tt(tun)*/){
                  mob->Wx=wpt->Xcoord;mob->Wy=wpt->Ycoord;
                  /* Free all icheck, ichecked, tcheck, and tchecked. */
                  for(;icheck!=NULL; isearch=icheck->Next,
                   free_intersearch(icheck),icheck=isearch);
                  for(;ichecked!=NULL; isearch=ichecked->Next,
                   free_intersearch(ichecked),ichecked=isearch);
                  for(;tcheck!=NULL; tsearch=tcheck->Next,
                   free_tunnelsearch(tcheck),tcheck=tsearch);
                  for(;tchecked!=NULL; tsearch=tchecked->Next,
                   free_tunnelsearch(tchecked),tchecked=tsearch);
                  return;
                }
                tsearch=new_tunnelsearch();
                tsearch->This=tun;tsearch->Waypoint=wpt;
                tsearch->Next=tcheck;
                tcheck=tsearch;
             }
           }
           for(;tcheck!=NULL;
           tsearch=tcheck,tcheck=tcheck->Next,
           tsearch->Next=tchecked,tchecked=tsearch)
           {
             tun=tcheck->This;wpt=tcheck->Waypoint;
             for(inter=get_intersections(tun,msregion->Intersections);
              inter!=NULL;
              inter=get_intersections(tun,inter->Next)){
                /* Find if inter is in ichecked, if so skip to next inter*/
                p=0;
                for(isearch=ichecked;isearch!=NULL;isearch=isearch->Next)
                  if(isearch->This==inter) {p=1;continue;}
                if(p==1/*in_ichecked(inter)*/) continue;
                /* Nothing else to check? */
                isearch=new_intersearch();
                isearch->This=inter;isearch->Waypoint=wpt;
                isearch->Next=icheck;
                icheck=isearch;
             }
           }
           /* And lo! we *ought to* now have a brand new list of
           intersections to check */
         } /* We exhausted all possibilities and found nothing. */
         /* Pick some random intersection with some random tunnel
         in mtunnels[]? Or just crash!  */
         sprintf(strbuf,"Monster at %d,%d couldn't find path to %d,%d.",mx,my,tx,ty);
         bugger(REPTYPE_MAPGEN|REPTYPE_AI,LOGLEVEL_FOOBAR,strbuf);
         /* We no longer consider crashing here. bugger() should
         alert the player when severity is LOGLEVEL_FOOBAR. */
         break;
       case 1: /* Rooms. Should be uncomplicated */
         /* Generally, set waypoint to tx,ty. Except, we could
         also set waypoints that are at some random point roughly
         in the middle of the room, or between middle of room
         and current location, etc...
         Or even towards any obvious items in the room?
         Then mob can act when it steps on them. */
         mob->Wx=tx;mob->Wy=ty;
         break;
       case 2: /* Doorways. Ah. Depends on whether target is
         in same region or same room... This should really
         have been converted to case 0 or case 1 as appropriate */
         crash(INCONSISTENT);
         break;
    }
}

tunnellist_t *get_tunnels(tunnellist_t *head,int x,int y)
{
    for(;head!=NULL;head=head->Nexttunnel)
    {
       if(head->Xstart<=x &&head->Xend>=x
       && head->Ystart<=y &&head->Yend>=y)
          /* this is good! */
          return(head);
    }
    return(NULL);
}

intersection_t *get_intersections(tunnellist_t *section,intersection_t *head)
{
    int tx1=section->Xstart,tx2=section->Xend,
    ty1=section->Ystart,ty2=section->Yend,
    x,y;
    for(;head!=NULL;head=head->Next)
    {
       x=head->Xcoord;y=head->Ycoord;
       if(tx1<=x &&tx2>=x
       && ty1<=y &&ty2>=y) return(head);
    }
    return(NULL);
}

tunnelsearch_t *new_tunnelsearch()
{
    tunnelsearch_t *new;
    /* We have an array of ~20 tunnelsearch_t elements, and
    we just return pointers to one of these. If they've all
    been allocated, counting_malloc() one, but set with
    a flag to say its been mallocced and wants to be freed.*/
    if(tscount<20) {new=ts_array+tscount;tscount++;return(new);}
    new=(tunnelsearch_t *)counting_malloc(MEM_TSEARCH,sizeof(tunnelsearch_t));
    if(new==NULL) crash(NO_MEM);
    new->Freeit=1; /* You must free this later! */
    tscount++;
    return(new);
}

intersearch_t *new_intersearch()
{
    intersearch_t *new;
    if(iscount<20) {new=is_array+iscount;iscount++;return(new);}
    new=(intersearch_t *)counting_malloc(MEM_ISEARCH,sizeof(intersearch_t));
    if(new==NULL) crash(NO_MEM);
    new->Freeit=1;
    iscount++;
    return(new);
}

void free_tunnelsearch(tunnelsearch_t *old)
{
    tscount--;
    if(old->Freeit)counting_free(MEM_TSEARCH,old);
}

void free_intersearch(intersearch_t *old)
{
    iscount--;
    if(old->Freeit)counting_free(MEM_ISEARCH,old);
}

intersection_t *nearest_intersection(int mx,int my)
{
    tunnellist_t *tun;
    intersection_t *inter;
    intersection_t *best=NULL;
    region_t *msregion=find_region(mx,my);
    intersection_t *head=msregion->Intersections;
    tunnellist_t *mtunnels[4]; /* Which tunnels are found here */
    int mt_tot=0;
    int dx,dy,p;
    int md=MAPHGT+MAPWID; /* Set to absolute max */
    int interd;
    for(p=0;p<4;p++)mtunnels[p]=NULL;
    for(tun=get_tunnels(msregion->Firsttunnel,mx,my);tun!=NULL;tun=get_tunnels(tun->Nexttunnel,mx,my))
    {
        if(mt_tot>3) crash(INCONSISTENT);
        mtunnels[mt_tot]=tun;mt_tot++;
    }
    /* Now find all intersections that fall in MT, and get one
    with min distance. */
    for(p=0;p<mt_tot;p++) {
       tun=mtunnels[p];
       for(inter=get_intersections(tun,head);
       inter!=NULL;inter=get_intersections(tun,inter->Next)) {
          dx=abs(mx-inter->Xcoord);dy=abs(my-inter->Ycoord);
          interd=dx+dy; /* Distance from mx,my to intersection */
          if(interd<md) {md=interd;best=inter;}
       }
    }
/*  if(best==NULL) crash(INCONSISTENT); We couldn't find *any*?*/
    return(best);
}

