/* ==== aicache.c ==== */
/* utilities pertaining to the cache system for the AI */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include "C-foogrid.h"
#include "porrog_defs.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h"
#include "ai_t.h"

/* protos */
#include "memutils.h"
#include "ai_p.h"
#include "crash.h"


aicache_tunnelent_t *aicache_regcache(monster_t *mob,region_t *region, int buf)
{
    aicache_memory_t *mmem=mob->Memory;
/*  aicache_tunnelent_t *buf=mob->Memory->Tunnelbuffer; */
    aicache_tunnelent_t *scan=mmem->Tunnelcache;
/*  if(buf!=NULL&&buf->Tunnel==region) return(buf); */
    if (buf&&mmem->Tunnelbuf_flag) return(mmem->Tunnelbuffer); /* Try buffer */
    mmem->Tunnelbuf_flag=0; /* Hmm, this is done elsewhere but doesnt always
      seem to get called. I'm not happy about this. */
    for(;scan!=NULL;scan=scan->Next)
    {
        if (scan->Tunnel==region) return(scan);
    }
    /* This should now NEVER happen, except when mob is first
    created, when aicache_roomcache() wont have been called
    before ai_acquire_target() :P */
    return(NULL);
}

aicache_rooment_t  *aicache_roomcache(monster_t *mob,int roomno,int buf)
{
    aicache_memory_t *mmem=mob->Memory;
/*  aicache_rooment_t *buf=mob->Memory->Roombuffer; */
    aicache_rooment_t *scan=mmem->Roomcache;
/*  if(buf!=NULL&&buf->Roomno) return(buf); */
    if (buf&&mmem->Roombuf_flag) return(mmem->Roombuffer); /* Try buffer*/
    mmem->Roombuf_flag=0; /* See comments in aicache_regcache() */
    for(;scan!=NULL;scan=scan->Next)
    {
        if (scan->Roomno==roomno) return(scan);
    }
    return(NULL);
}

/* Similar to the functions for retrieving cache entries, but these
attempt to place them (and set them up) as well.
NB- DOESNT detect if the entry is already there!!! You should have
called aicache_<whatever>cache() first to check this, and only
call these when that returned a NULL. */
aicache_tunnelent_t *aicache_trycacheregion(monster_t *mob,region_t *region)
{
    aicache_tunnelent_t *list=mob->Memory->Tunnelcache;
    aicache_tunnelent_t *scan=list;
    aicache_tunnelent_t *chosen=NULL;
    entrancelist_t *escan;
    int count=0; /* How many entries already exist (memory starts empty) */
    int lowestorder=100,thisorder=0;
    long oldest=0; /* The oldest timestamp found in the lowest order entries*/

    /* Count cache entries and find lowest order one */
    for(;scan!=NULL&&( (lowestorder=(scan->Order<lowestorder)?scan->Order:lowestorder),1);scan=scan->Next,count++);

    /* Now, find order of region- If cache isn't full, we'll need it to
    fill in details. If it is full, we'll need it to see if we can
    replace anything. */
    for(escan=region->Firstentrance;escan!=NULL;escan=escan->Next,thisorder++);

    /* Now, if cache isn't yet full, we will cache the region regardless */
    if (count<3) {
       chosen=(aicache_tunnelent_t *)counting_malloc(MEM_TUNMEM,sizeof(aicache_tunnelent_t));
       if (chosen==NULL) crash(NO_MEM);
       if(count==0) { /* Totally empty */
          mob->Memory->Tunnelcache=chosen;
       } else {
          for(scan=list;scan->Next!=NULL;scan=scan->Next);
          scan->Next=chosen;
       }
       chosen->Paths=NULL;
       chosen->Next=NULL; /* Because chosen is put at the end, right? */
       /* Er, wouldn't it be better to stick chosen at the start?? */
       /* We'll populate it after. */
    } else {
       /* Cache is full, but we might replace an existing element. */
       /* If any elements have lower order than this region, that
       is a suitable one. Otherwise, any elements with the same
       order as this region is suitable. Otherwise, consider regions
       of higher order with very old timestamps to be suitable
       (say, 1500 turns is equivalent to one door) */
       if (lowestorder<=thisorder) {
          for (scan=list;scan!=NULL;scan=scan->Next)
          {
              /* One of these (with lowestorder) is going to be
              replaced, only question is which: the oldest. */
              if(scan->Order==lowestorder)
                 if (scan->Age>oldest) {
                    oldest=scan->Age; chosen=scan;
                 }
              /* Righto */

          }
          /* chosen is now the element to replace. */
       } else { /* Didn't find one suitable. We might take age into
          account though. */

          ;  /* Or just not bother */
       }
    }
/*  if (NULL==chosen) return (NULL);
    NOPE! Now, if we couldn't place the location in the cache, we
    place it in a buffer entry instead. The buffer entry only lasts
    until the monster leaves the location, but stops it leaving
    through the entrance it came in, and also simplifies the
    ai_acquire_target() function! */
    if(NULL==chosen) {
      mob->Memory->Tunnelbuf_flag=1; /* Show it's valid */
      chosen=mob->Memory->Tunnelbuffer; /* It's already malloced */
    } else mob->Memory->Tunnelbuf_flag=0; /* Isn't valid */
    /* OK, so we got one. */
    /* First, free the doors list (if any). */
    if (chosen->Paths!=NULL) aicache_freedoorlist(chosen->Paths);
    chosen->Tunnel=region;
    chosen->Age=0;
    chosen->Order=thisorder;
    chosen->Paths=aicache_convertdoors(region->Firstentrance);
    /* That also sets their ages to 2 */
    return(chosen);
}

aicache_rooment_t *aicache_trycacheroom(monster_t *mob, int roomno)
{
    aicache_rooment_t *list=mob->Memory->Roomcache;
    aicache_rooment_t *scan=list;
    aicache_rooment_t *chosen=NULL;
    entrancelist_t *escan;
    int count=0; /* How many entries already exist (memory starts empty) */
    int lowestorder=100,thisorder=0;
    long oldest=0; /* The oldest timestamp found in the lowest order entries*/

    for(;scan!=NULL&&( (lowestorder=(scan->Order<lowestorder)?scan->Order:lowestorder),1);scan=scan->Next,count++);

    /* Find order of this room */
    for(escan=tls->Room[roomno].Firstentrance;escan!=NULL;escan=escan->Next,thisorder++);

    if (count<3) {
       chosen=(aicache_rooment_t *)counting_malloc(MEM_ROOMMEM,sizeof(aicache_rooment_t));
       if(chosen==NULL) crash(NO_MEM);
       if(count==0) { /* Totally empty */
          mob->Memory->Roomcache=chosen;
       } else {
          for(scan=list;scan->Next!=NULL;scan=scan->Next);
          scan->Next=chosen;
       }
       chosen->Paths=NULL;
       chosen->Next=NULL;
    } else {
       /* Cache is full, but consider replacing */
       if (lowestorder<=thisorder) {
          for (scan=list;scan!=NULL;scan=scan->Next)
          {
              /* One of these (with lowestorder) is going to be
              replaced, only question is which: the oldest. */
              if(scan->Order==lowestorder)
                 if (scan->Age>oldest) {
                    oldest=scan->Age; chosen=scan;
                 }
          }
          /* Should now have something chosen */
       } else {
          /* We could consider elements with higher order,
          if they're very old. But not yet. */
          ;
       }
    }
/*  if (NULL==chosen) return(NULL);
    As explained in aicache_trycachetunnel(), we now store entries
    we couldn't cache into a buffer so they aren't lost immediately */
    if (NULL==chosen) {
      mob->Memory->Roombuf_flag=1;
      chosen=mob->Memory->Roombuffer;
    } else mob->Memory->Roombuf_flag=0; /* Not valid */
    /* We got one. */
    if (chosen->Paths!=NULL) aicache_freedoorlist(chosen->Paths);
    chosen->Roomno=roomno;
    chosen->Age=0;
    chosen->Order=thisorder;
    chosen->Paths=aicache_convertdoors(tls->Room[roomno].Firstentrance);
    /* And make them age 2 */
    return(chosen);
}

void aicache_freerooments(aicache_rooment_t *head)
{
    aicache_rooment_t *next;
    for(;head!=NULL; next=head->Next,aicache_freedoorlist(head->Paths),counting_free(MEM_ROOMMEM,head),head=next)
    ;
}

void aicache_freetunnelents(aicache_tunnelent_t *head)
{
    aicache_tunnelent_t *next;
    for(;head!=NULL; next=head->Next,aicache_freedoorlist(head->Paths),counting_free(MEM_TUNMEM,head),head=next)
    ;
}

void aicache_freedoorlist(aicache_doorlist_t *list)
{
    aicache_doorlist_t *next;
    /* counting_free(this),this=next,   or
       next=next,counting_free(this),this=next */
    for(;list!=NULL;next=list->Next,counting_free(MEM_DOORMEM,list),list=next)
    {
        counting_free(MEM_DOORMEM,list->Thisdoor);
    }
}

void aicache_tryresetdoor(monster_t *mob,aicache_doorlist_t *list)
{
    int mx=mob->X,my=mob->Y;
    aicache_doorent_t *door;
    for (;list!=NULL;list=list->Next)
    {
        door=list->Thisdoor;
        if(door->Xcoord==mx && door->Ycoord==my) {door->Age=0;return;}
    }
    crash(INCONSISTENT); /* We shouldn't have been called unless the
    mob was standing on a door that was in the cache entry */
}

aicache_doorlist_t *aicache_convertdoors(entrancelist_t *list)
{
    entrance_t *source;
    aicache_doorent_t *target;
    aicache_doorlist_t *head=NULL;
    aicache_doorlist_t *chunk;
    for (;list!=NULL;list=list->Next)
    {
        source=list->Entrance;
        target=(aicache_doorent_t *)counting_malloc(MEM_DOORMEM,sizeof(aicache_doorent_t));
        if (target==NULL) crash(NO_MEM);
        target->Xcoord=source->Xcoord;
        target->Ycoord=source->Ycoord;
        target->Age=2; /* So when we reset the door we're at, it really does
          end up newer than the others */
        chunk=(aicache_doorlist_t *)counting_malloc(MEM_DOORMEM,sizeof(aicache_doorlist_t));
        if (chunk==NULL) crash(NO_MEM);
        chunk->Next=head;
        chunk->Thisdoor=target;
        head=chunk;
    }
    return(head);
}

void aicache_updatetimes(monster_t *mob)
{
    /* Should we also update the times on buffer entries?
    Prolly not. Updating the times is only useful for
    when you come back to the place, and buffer entries
    aren't stored, so there. */
    aicache_doorlist_t *dlist;
    aicache_rooment_t *rooment=mob->Memory->Roomcache;
    aicache_tunnelent_t *regent=mob->Memory->Tunnelcache;
    /* First update times on roomcaches (and their doors) */
    for (;rooment!=NULL;rooment=rooment->Next)
    {
        rooment->Age++;
        for(dlist=rooment->Paths;dlist!=NULL;dlist=dlist->Next)
           /* Consider only increasing ages of doors
           that are below a certain number, so doors that
           haven't been passed for a very long time end
           up with equal age. Not only does it prevent
           potential overflow, but also makes it possible
           to change the order in which we go through doors
           (which otherwise gets set to whatever order was
           first chosen) */
           dlist->Thisdoor->Age++;
    }
    /* Now update times on tunnelcaches (and their doors) */
    for (;regent!=NULL;regent=regent->Next)
    {
        regent->Age++;
        for(dlist=regent->Paths;dlist!=NULL;dlist=dlist->Next)
           dlist->Thisdoor->Age++;
    }
}

