/* ==== combat.c ==== */
/* Uh... what'll we do here? */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "porrog_defs.h"
#include "C-foogrid.h"
#include "item_t.h"
#include "crash.h"
#include "structures.h"
#include "mapgen_types.h"
#include "player_t.h"
#include "dXrng.h"
/* protos */
#include "monster.h"
#include "ui.h"
#include "los.h"
#include "combat.h"
#ifdef AUTOPILOT
#include   "mg_utils.h"
#endif


/* Don't need to state map any more ! */
void p_hit_m(monster_t *target)
{
    /* For now, we just assume player always hits
    and removes 1 HP from target */
    target->Var_stats.Hp--;
    sprintf(strbuf,"You hit the %s.",target->Mobtype->Name);
    show_messages(strbuf);
    if(target->Var_stats.Hp<0)
    {
       sprintf(strbuf,"You kill the %s!",target->Mobtype->Name);
       show_messages(strbuf);
       m_die(tls, target); /* should be in monster.c */
    }
}

void m_hit_p(monster_t *attacker)
{
#ifdef AUTOPILOT
    int r;
#endif
    int dmg=2; /* should really be dependent on weapon, etc */
    int tpos,apos,acov, aprot;
    item_t *aitem;
    /* Originally, just assume attacker always hits, removing 1 HP */
    /* Haha! Now we've got more advanced view of combat. */
    /* First see if mob hits you. */
    if (roll_dX(d20set[0])<player.Var_stats.Ev) { /* Missed! */
       sprintf(strbuf,"The %s misses you.",attacker->Mobtype->Name);
       show_messages(strbuf);
       return; 
    }
    /* The mob has hit. Does it hit your armour or *you*? */
    tpos=(10*roll_dX(d10set[0]))+roll_dX(d10set[1]);
    if (tpos<(player.Var_stats.Cov)) { /* hit armour */
       /* Work out *what* it hit... */
       for(apos=EQUIP_SUIT;apos!=EQUIP_NONE;tpos-=acov,apos++,apos=((apos==EQUIP_RING1)?EQUIP_NECKWEAR:apos))
       {
          /* find coverage of this item */
          aitem=player.Equipment[apos];
          if(aitem==NULL) {acov=0; continue;}
          acov=aitem->Itemtype->Typedata.Arm.Cov;
          if(tpos<acov) break; /* This is the item blocking */
       }
       if (apos==EQUIP_NONE) crash(FUNCTION_ERROR);
       aprot=aitem->Itemtype->Typedata.Arm.Prot;
       if (aprot>=dmg) { /* Damage nullified */
          sprintf(strbuf,"The %s hits you, but your %s completely blocks it.",attacker->Mobtype->Name,aitem->Itemtype->Name);
          dmg=0;
       } else { /* Reduced the damage */
          sprintf(strbuf,"The %s hits you (your %s absorbs some of the blow).",attacker->Mobtype->Name,aitem->Itemtype->Name);
          dmg-=aprot;
       }
    } else { /* bypasses armour */
       sprintf(strbuf,"The %s hits you.",attacker->Mobtype->Name);
    }
/* player.Var_stats.Hp-=((roll_dX(d6set[0])*(100-player.Var_stats.Cov))/200);*/
    player.Var_stats.Hp-=dmg;
    show_messages(strbuf);
    if(player.Var_stats.Hp<0)
    {
#ifndef AUTOPILOT
       sprintf(strbuf,"The %s kills you!",attacker->Mobtype->Name);
       show_messages(strbuf);
       quit=2; /* 2=dead */
#else
       /* Don't quit, just put the player somewhere else instead */
       player.Var_stats.Hp=player.Var_stats.MaxHp;
       r=pick_room(tls->Totalrooms);
       player.X=tls->Room[r].Xcentre;
       player.Y=tls->Room[r].Ycentre;
#endif
    }
}

void m_hit_m(monster_t *att, monster_t *target)
{
    /* For now, we just assume player always hits
    and removes 1 HP from target */
    target->Var_stats.Hp--;
#ifndef AUTOPILOT
    if (in_los(player.X,player.Y,att->X,att->Y)) {
#else
    if (in_los(tls->Moblist->X,tls->Moblist->Y,att->X,att->Y)) {
#endif
      sprintf(strbuf,"The %s hits the %s.",att->Mobtype->Name,target->Mobtype->Name);
      show_messages(strbuf);
    }
    if(target->Var_stats.Hp<0)
    {
#ifndef AUTOPILOT
       if (in_los(player.X,player.Y,att->X,att->Y)) {
#else
       if (in_los(tls->Moblist->X,tls->Moblist->Y,att->X,att->Y)) {
#endif
         sprintf(strbuf,"The %s kills the %s!",att->Mobtype->Name,target->Mobtype->Name);
         /* Would be handy to log this for debug stuff */
         show_messages(strbuf);
       }
       m_die(tls, target); /* should be in monster.c */
    }
}
