/* ==== crash.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* #include <stdio.h> included by ui.h???? */
#include <ncurses.h> /* Because we don't do it properly yet. */
#include "porrog_defs.h" /* EVERYTHING needs this now.*/
#include "crash.h"
#include "item_t.h"
#include "structures.h" /* needed for item_t def*/
#include "ui.h" /*We're doing it through curses now.*/
#include "memutils.h"


void crash(enum crashtype reason)
{
/*    fprintf(stderr,"\nDOH!\n Porrog has encountered an error!\n");
*/
    /* Should we do this here? */
    dump_buglog();
    switch (reason)
    {
        case NO_MEM:
           msg_window("Couldn't allocate more memory!");
           /*Either the game has a memory leak, or\n your machine doesn't have enough memory installed.\n");
           */
           break;
        case INVALID_DIE_TYPE:
           msg_window("Tried to use an invalid die type.");
           /*Our RNG only supports D3, D4, D5, D6, D8, D10, D12, and D20 dice.\n");
           */
           break;
        case INVALID_ENVIRONMENT:
           msg_window("Porrog is unable to start in the environment supplied.");
           /*So, it didn't.\n");
           */
           break;
        case OUT_OF_BOUNDS:
           msg_window("Some thing was accessed outside its proper bounds.");
           /*This is a bug.\n");
           */
           break;
        case INCONSISTENT:
           msg_window("Some unknown, earlier error left inconsistent data we cant deal with.");
           break;
        case SCREEN_ERROR:
           msg_window("An error was encountered accessing the screen.");
           break;
        case IO_ERROR:
           msg_window("An IO error occurred. Somehow.");
           break;
        case FUNCTION_ERROR:
           msg_window("A specific function has gone wrong somehow.");
           break;
/*        default: */
           msg_window("The type of error is unknown! This is an error in itself!!!");
           break;
    }
    /*
    fprintf(stderr,"For this reason, Porrog has quit.\n");
    fprintf(stderr,"Well, unless the crash was down to you (such as invalid starting parameters),\nyou ought to let the author know.\nProbably.\n");
    */
/*    getch();  Wait for a key. */
    for(;;);
    endwin();
    exit(1);
}

void bugger(unsigned short flags, buglevel_t severity, char *bug)
{
    if((severity>LOGLEVEL_DEBUG)||(flags&&watchflags)) {
       /* This debug message is worth logging */
       if (bugger_ringbuffer[brb_endidx]!=NULL) counting_free(MEM_STRING,bugger_ringbuffer[brb_endidx]);
       bugger_ringbuffer[brb_endidx]=counting_malloc(MEM_STRING,sizeof(char)*(1+strlen(bug)));
       strcpy(bugger_ringbuffer[brb_endidx],bug);
       brb_endidx++;if(brb_endidx==200) brb_endidx=0;
       if (brb_endidx==brb_startidx) { /* Wrapped */
          brb_startidx++; if (brb_startidx==200) brb_startidx=0;
       }
       /* Logged. Consider also alerting player, if severity is
       LOGLEVEL_FOOBAR, or if game being run in DEBUG mode and
       severity is LOGLEVEL_WARNING. If running in DEBUG mode,
       player will be able to read the debug logs during the
       game, whilst outside of DEBUG mode, what remains of the
       ring buffer are just dumped to a file at the end of the
       game, and can't be read before that. */
       if(severity==LOGLEVEL_FOOBAR)show_messages("*BUG LOGGED*");
    }
}

void dump_buglog()
{
    int rbi;
    FILE *bugsfile=fopen("porrogbugs.log","wt"); /* text file, overwrite */
    if(bugsfile==NULL) {
       fprintf(stderr,"dump_buglog(): Couldn't open bugs log file to write to! THATS BAD!\n"); return;
    }

    for(rbi=brb_startidx;;rbi++) {
       if (rbi==200) rbi=0;
       if (rbi==brb_endidx) break; /* endidx is never full. */
       if (bugger_ringbuffer[rbi]==NULL) {
          fprintf(stderr,"NULL pointer found in buglog\n");
          break; /* will this get called? */
       }
       fputs(bugger_ringbuffer[rbi],bugsfile);
       fputc('\n',bugsfile);
    }
    if(fclose(bugsfile)==EOF) fprintf(stderr,"dump_buglog(): Error closing bugs log file! THATS BAD!\n");
}

void free_buglog()
{
    int rbi;
    for(rbi=brb_startidx;rbi!=brb_endidx;rbi++) {
       if (rbi==200) rbi=0;
       if (bugger_ringbuffer[rbi]==NULL) {
          fprintf(stderr,"free_buglog(): NULL pointer found in buglog.\n");
          break;
       }
       counting_free(MEM_STRING,bugger_ringbuffer[rbi]);
    }
    if(bugger_ringbuffer[rbi]!=NULL) counting_free(MEM_STRING,bugger_ringbuffer[rbi]);
    /* Check this is good. */
}

/* --- end of you know, stuff --- */
