/* ==== dXrng.c ==== */
/* An RNG for Porrog, a roguelike game */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include "porrog_defs.h"
#include "crash.h"
#include "dXrng.h"

/* Putting global arrays of dice in globals.c now. */
/* We could use d3s, or d5s. roll_dX USED to support
   these... But taking the % operators out (to replace
   them with conditionals, which are quicker) would
   probably stop these from working??? Must check this. */

/* The tables of shuffled dXs, chopped into 3 entry long sections */
byte d4Tab[8*3];   /* WAS  [8][3] */
byte d6Tab[12*3];  /* Was [12][3] */
byte d8Tab[16*3];  /* Was [16][3] */
byte d10Tab[20*3]; /* Was [20][3] */
byte d12Tab[24*3]; /* Was [24][3] */
byte d20Tab[40*3]; /* Was [40][3] */

/* The lists of section indices, that describe the sequences */
byte d4SeqList[NUMSEQS*8];   /* Was  [NUMSEQS][8] */
byte d6SeqList[NUMSEQS*12];  /* Was [NUMSEQS][12] */
byte d8SeqList[NUMSEQS*16];  /* Was [NUMSEQS][16] */
byte d10SeqList[NUMSEQS*20]; /* Was [NUMSEQS][20] */
byte d12SeqList[NUMSEQS*24]; /* Was [NUMSEQS][24] */
byte d20SeqList[NUMSEQS*40]; /* Was [NUMSEQS][40] */


/* WHY are these still here? I think this is from before it
   dawned on me to put such data in the roller structures,
   and mebbe I forgot to remove em here... */
byte primaryd4ListNumber; /* Initialised at startup */
byte primaryd4ListPointer=0; /* Pointer into a list */
byte primaryd4SectionPointer=0; /* Pointer to an entry within a section
                                  of a dX table*/

/* Ditto for d6, d8, d10, d12, d20. We would also want equivalents for
 secondary dice. But there could be lots! So really, these should be
 static in a function rollNewPrimaryd4();*/
/* Ah, it's now encompassed in dXrng_roller. OBLITERATE ABOVE NONSENSE!*/

/* Fills the dXTab arrays and shuffles them. */
void initialise_dXrng()
{
   int  pass,maxn;
   int s1Sec, s2Sec;
   /* int s1Ent,s2Ent; Do we not use these things? */
   byte entry;
   int *listcomp;

   /* First d4: 6 sets of 4 numbers=24 */
   for (pass=0;pass<24;pass++)
     d4Tab[pass]=(byte)pass%4;
   for (pass=0;pass<12*4;pass++)
   {
     s1Sec=pass%24;
     /*s1Ent=(pass/8)%3; */
     /*s2Sec=dXrng_sysrandom(8);
     s2Ent=dXrng_sysrandom(3); */
     s2Sec=dXrng_sysrandom(24);
     /* entry=d4Tab[s1Sec][s1Ent];*/
     entry=d4Tab[s1Sec];
     d4Tab[s1Sec]=d4Tab[s2Sec];
     d4Tab[s2Sec]=entry;
     /* A *real* shuffler would prolly check that each entry
     was moved at least once. Oh yeah, it does now. */
   }
   /* Now d6: 6 sets of 6 numbers=36 */
   for (pass=0;pass<36;pass++)
     d6Tab[pass]=(byte)pass%6;
   for (pass=0;pass<12*6;pass++)
   {
     s1Sec=pass%36;
     /* s1Ent=(pass/12)%3;
     s2Sec=dXrng_sysrandom(12);
     s2Ent=dXrng_sysrandom(3); */
     s2Sec=dXrng_sysrandom(36);
     entry=d6Tab[s1Sec];
     d6Tab[s1Sec]=d6Tab[s2Sec];
     d6Tab[s2Sec]=entry;
   }

   /* Now d8: 6 sets of 8 numbers=48 */
   for (pass=0;pass<48;pass++)
     d8Tab[pass]=(byte)pass%8;
   for (pass=0;pass<12*8;pass++)
   {
     s1Sec=pass%48;
     s2Sec=dXrng_sysrandom(48);
     entry =       d8Tab[s1Sec];
     d8Tab[s1Sec]= d8Tab[s2Sec];
     d8Tab[s2Sec]= entry;
   }

   /* Now d10: 6 sets of 10 numbers=60 */
   for (pass=0;pass<60;pass++)
     d10Tab[pass]=(byte)pass%10;
   for (pass=0;pass<12*10;pass++)
   {
     s1Sec=pass%60;
     s2Sec=dXrng_sysrandom(60);
     entry =        d10Tab[s1Sec];
     d10Tab[s1Sec]= d10Tab[s2Sec];
     d10Tab[s2Sec]= entry;
   }

   /* Now d12: 6 sets of 12 numbers=72 */
   for (pass=0;pass<72;pass++)
     d12Tab[pass]=(byte)pass%12;
   for (pass=0;pass<12*12;pass++)
   {
     s1Sec=pass%72;
     s2Sec=dXrng_sysrandom(72);
     entry =        d12Tab[s1Sec];
     d12Tab[s1Sec]= d12Tab[s2Sec];
     d12Tab[s2Sec]= entry;
   }

   /* Now d20: 6 sets of 20 numbers=120 */
   for (pass=0;pass<120;pass++)
     d20Tab[pass]=(byte)pass%20;
   for (pass=0;pass<12*20;pass++)
   {
     s1Sec=pass%120;
     s2Sec=dXrng_sysrandom(120);
     entry =        d20Tab[s1Sec];
     d20Tab[s1Sec]= d20Tab[s2Sec];
     d20Tab[s2Sec]= entry;
   }

   /* Tables populated. Now shuffle. */
   /* We already have here, but perhaps should be in separate function. */

   /* Now we create the section sequence lists. */
   listcomp=(int *)malloc(40*sizeof(int));
   if (NULL==listcomp) crash(NO_MEM);

   for (pass=0;pass<NUMSEQS;pass++)
   {
      /* d4 lists */
      /* Generate list of full scale random numbers */
      for (s1Sec=0;s1Sec<4*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0); /*0 means full range*/
      for (s2Sec=0;s2Sec<4*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<4*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d4SeqList[pass][entry]=s2Sec; */
         d4SeqList[entry+(4*2*pass)]=s2Sec;
      }
      /* d6 lists */
      /* Generate list of full scale random numbers */
      for (s1Sec=0;s1Sec<6*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0);
      for (s2Sec=0;s2Sec<6*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<6*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d6SeqList[pass][entry]=s2Sec; */
         d6SeqList[entry+(6*2*pass)]=s2Sec;
      }
      /* d8 lists */
      /* Generate list of full scale random numbers */
      for (s1Sec=0;s1Sec<8*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0);
      for (s2Sec=0;s2Sec<8*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<8*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d8SeqList[pass][entry]=s2Sec; */
         d8SeqList[entry+(8*2*pass)]=s2Sec;
      }
      /* d10 lists */
      /* Generate list of random numbers */
      for (s1Sec=0;s1Sec<10*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0);
      for (s2Sec=0;s2Sec<10*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<10*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d10SeqList[pass][entry]=s2Sec; */
         d10SeqList[entry+(10*2*pass)]=s2Sec;
      }
      /* d12 lists */
      /* Generate list of random numbers */
      for (s1Sec=0;s1Sec<12*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0);
      for (s2Sec=0;s2Sec<12*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<12*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d12SeqList[pass][entry]=s2Sec; */
         d12SeqList[entry+(12*2*pass)]=s2Sec;
      }
      /* d20 lists */
      /* Generate list of random numbers */
      for (s1Sec=0;s1Sec<20*2;s1Sec++)
         listcomp[s1Sec]=dXrng_sysrandom(0);
      for (s2Sec=0;s2Sec<20*2;s2Sec++)
      {
         maxn=-1;
         for (s1Sec=0;s1Sec<20*2;s1Sec++)
            if (listcomp[s1Sec]>maxn)
            {
               maxn=listcomp[s1Sec];
               entry=(byte)s1Sec;
            }
         listcomp[entry]=-1;
         /* d20SeqList[pass][entry]=s2Sec; */
         d20SeqList[entry+(20*2*pass)]=s2Sec;
      }
   }
   free(listcomp);
   /* Return nothing. */
}

void make_dXen()
{
   int count;
   /* d4s */
   for (count=0;count<8;count++)
     d4set[count]=new_dX(4);
     /* Hey, wouldn't it be much better to have something like
     new_dX(d4set+count,4); such that new_dX() didn't have to
     use malloc? */
   /* d6s */
   for (count=0;count<5;count++)
     d6set[count]=new_dX(6);
   /* d8s */
   for (count=0;count<4;count++)
     d8set[count]=new_dX(8);
   /* d10s */
   for (count=0;count<3;count++)
     d10set[count]=new_dX(10);
   /* d12s */
   for (count=0;count<3;count++)
     d12set[count]=new_dX(12);
   /* d20s */
   for (count=0;count<2;count++)
     d20set[count]=new_dX(20);
   /* That's it. */
}


/* Creates a new die roller. NB- This also supports d3s and d5s, which
   are simulated by d6s and d10s. NB2- Remember that this allocates memory. */
dXrng_roller *new_dX(int sides)
{
    dXrng_roller *die_ptr;
    if (sides<3 ||sides>20) crash(INVALID_DIE_TYPE);
    die_ptr=(dXrng_roller *) malloc(sizeof(dXrng_roller));
    if (NULL==die_ptr) crash(NO_MEM);
    die_ptr->dXType=sides;
    switch (sides)
    {
        case 4:
           die_ptr->N_Num=8;
           die_ptr->dXTab=d4Tab;
           die_ptr->SeqList=d4SeqList;
           break;
        case 3:
        case 6:
           die_ptr->N_Num=12;
           die_ptr->dXTab=d6Tab;
           die_ptr->SeqList=d6SeqList;
           break;
        case 8:
           die_ptr->N_Num=16;
           die_ptr->dXTab=d8Tab;
           die_ptr->SeqList=d8SeqList;
           break;
        case 5:
        case 10:
           die_ptr->N_Num=20;
           die_ptr->dXTab=d10Tab;
           die_ptr->SeqList=d10SeqList;
           break;
        case 12:
           die_ptr->N_Num=24;
           die_ptr->dXTab=d12Tab;
           die_ptr->SeqList=d12SeqList;
           break;
        case 20:
           die_ptr->N_Num=40;
           die_ptr->dXTab=d20Tab;
           die_ptr->SeqList=d20SeqList;
           break;
        default:
           crash(INVALID_DIE_TYPE);
    }
    die_ptr->ListPtr=0; /*Start of the list*/
    die_ptr->SecPtr=0;  /*First entry of a section*/
    die_ptr->ListNum=dXrng_sysrandom(NUMSEQS);  /*The sequence list to follow*/
    die_ptr->S_Offset=dXrng_sysrandom(die_ptr->N_Num); /*To modulate section*/
    die_ptr->E_Offset=dXrng_sysrandom(3);     /*To modulate entry no*/
    die_ptr->R_Offset=dXrng_sysrandom(sides); /*To modulate the roll*/
    /* Er... That's it?*/
    return die_ptr;
}

int dXrng_sysrandom(int range)
{
    if (0==range) return random(); /* indicates full range */
    return (random()%range); /*That's not really good enough, but OK for now*/
}

int roll_dX(dXrng_roller *die)
{
    byte ListNum,ListPtr,SecPtr, N_Num,range;
    byte *dXTab;
    byte *SeqList;
    int roll;
    int section, entry;

    N_Num=die->N_Num;
    range=die->dXType;
    ListNum=die->ListNum;
    ListPtr=die->ListPtr;
    SecPtr=die->SecPtr;

    dXTab=die->dXTab;
    SeqList=die->SeqList;

    /*The die already points to a valid roll.*/

    /* section=(die->S_Offset + SeqList[ListNum][ListPtr] )%N_Num; */
    /*section=(die->S_Offset + *((SeqList+(N_Num*ListNum))+ListPtr))%N_Num; */
    section=(die->S_Offset + *((SeqList+(N_Num*ListNum))+ListPtr));
    section=(section>=N_Num) ? section-N_Num : section; /* Faster? */

    /*entry=  (die->E_Offset + SecPtr) %3; */
    entry=(die->E_Offset + SecPtr);
    entry=(entry>2) ? entry-3 : entry; /* Faster */
    /* roll=   (die->R_Offset + dXTab[section][entry] )%range; */
    /*roll= (die->R_Offset + *(dXTab+(3*section)+entry) )%range; */
    roll= (die->R_Offset + *(dXTab+(3*section)+entry) );
    roll=(roll>=range) ? roll-range : roll; /* Hopefully faster...*/

    /* How simple is that!?!*/
    SecPtr++;
    if (SecPtr>2)
    {
        SecPtr=0;
        ListPtr++;
        if (ListPtr>=N_Num)
        {
           /*
           ListPtr=0;
           die->ListNum=(ListNum+1+dXrng_sysrandom(NUMSEQS-1))%NUMSEQS; */

           ListPtr=(ListNum+1+dXrng_sysrandom(NUMSEQS-1)); /* Temp var! */
           ListPtr=(ListPtr>=NUMSEQS)? ListPtr-NUMSEQS : ListPtr;
           die->ListNum=ListPtr;/*Didn't want to manipulate die->ListNum as rw*/
           /* We don't want it to be *the same* list.*/
           ListPtr=0; /* Start at the start of the new list. */
           /* Note also that here we've made it clear that the previous
           assignments to ListPtr used it as a temporary variable */

           /*Generate new offsets*/
           die->S_Offset=dXrng_sysrandom(N_Num);
           die->E_Offset=dXrng_sysrandom(3);
           die->R_Offset=dXrng_sysrandom(range);
        }
        die->ListPtr=ListPtr;
    }
    die->SecPtr=SecPtr;
    /* Die is now updated for the next roll.*/
    /* "roll" is range 0->X-1. We could either add 1 here for a proper dX,
       or add it in the caller, or have a flag to determine type, or
       have 2 separate functions, to support each type. */
    return roll;
}

