/* ==== i_init.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */




/* This initialises the set of item types. Still not sure if such
   data should be inside the .c source or a header... No, it
   would make no sense to put it in a header, as it would only
   be #included once. */

/* Need helper functions that do more than just "new_object_type",
 and define type-specific things */

#include <stddef.h>
#include "porrog_defs.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "C-foogrid.h" /* dammit we HAVE to separate protos from typedefs. */

/* Protos */
#include "item_p.h"
#include "memutils.h"

#define DEF_WEAP(name) push_itype(IG_WEAPON,name);
#define DEF_BOW(name) push_itype(IG_RANGED,name);
#define DEF_MISSILE(name) push_itype(IG_MISSILE,name);

/* Coverage percentages:
  SUIT=40%.   SHIELD=15%,   HEADGEAR=15%,
  GLOVES=8%   FOOTWEAR=20%, NECKWEAR=2%.
  Rings have no coverage (same place as gloves). */
/* Armour items */
#define DEF_SUIT(td,name,cov,prot) td=push_itype(IG_SUIT,name),td->Typedata.Arm.Cov=(cov*0.4), td->Typedata.Arm.Prot=prot;
#define DEF_SHIELD(td,name,cov,prot) td=push_itype(IG_SHIELD,name),td->Typedata.Arm.Cov=(cov*0.15), td->Typedata.Arm.Prot=prot;
#define DEF_HEADGEAR(td,name,cov,prot) td=push_itype(IG_HEADGEAR,name),td->Typedata.Arm.Cov=(cov*0.15), td->Typedata.Arm.Prot=prot;
#define DEF_GLOVES(td,name,cov,prot) td=push_itype(IG_GLOVES,name),td->Typedata.Arm.Cov=(cov*0.08), td->Typedata.Arm.Prot=prot;
#define DEF_FOOTWEAR(td,name,cov,prot) td=push_itype(IG_FOOTWEAR,name),td->Typedata.Arm.Cov=(cov*0.2), td->Typedata.Arm.Prot=prot;

/* Jewellery */
#define DEF_RING(name) push_itype(IG_RING,name);
#define DEF_NECKWEAR(td,name,cov,prot) td=push_itype(IG_NECKWEAR,name), td->Typedata.Arm.Cov=(cov*0.02), td->Typedata.Arm.Prot=prot;

/* Readables */
#define DEF_SCROLL(name)  push_itype(IG_SCROLL,name);
#define DEF_BOOK(name) push_itype(IG_BOOK,name);

/* Comestibles */
#define DEF_FOOD(name) push_itype(IG_FOOD,name);
#define DEF_CONDIMENT(name) push_itype(IG_CONDIMENT,name);
#define DEF_POTION(name)  push_itype(IG_POTION,name);
/* Zappables */
#define DEF_WAND(name)  push_itype(IG_WAND,name);
#define DEF_ROD(name)   push_itype(IG_ROD,name);
#define DEF_TOOL(name)  push_itype(IG_TOOL,name);
#define DEF_CONTAINER(name,opens,wlim,iclass,extype)  td=push_itype(IG_CONTAINER,name),td->Typedata.Container.Opens=opens,td->Typedata.Container.Wlim=wlim,td->Typedata.Container.AcceptsClass=iclass,td->Typedata.Container.AcceptsSpec=extype;
#define DEF_STONE(name)  push_itype(IG_STONE,name);
#define DEF_COINAGE(name)  push_itype(IG_COINAGE,name);

void i_init()
{
    itemtype_t *td;
    phonyitem=(item_t *)counting_malloc(MEM_ITEM,sizeof(item_t));
    phonyitem->Itemtype=NULL; /* Number element used to
    describe whether to unwield or unready */
#ifdef MARK_INTERSECTIONS
    /* This is freed by free_everything() */
    corner_marker=(itemtype_t *)counting_malloc(MEM_ITYPE,sizeof(itemtype_t));
    new_itemtype(corner_marker,'X',"Intersection marker",IG_STONE);
#endif
    /* Wieldables/wearables */
    DEF_WEAP("sword");
    DEF_WEAP("axe");
    DEF_WEAP("spear");
    DEF_WEAP("club");
    DEF_WEAP("knife");
    DEF_WEAP("claw hammer");
    DEF_WEAP("mallet");
    DEF_WEAP("sledgehammer");
    DEF_WEAP("exacto knife");
    DEF_WEAP("penknife");
    DEF_WEAP("hatchet");
    DEF_WEAP("cleaver");
    DEF_WEAP("whip");

    DEF_BOW("bow");
    DEF_BOW("crossbow");

    DEF_MISSILE("arrow");
    DEF_MISSILE("bolt");

    DEF_SUIT(td,"pinstripe suit",99,1);
    DEF_SUIT(td,"boiler suit",90,2);
    DEF_SUIT(td,"gimp suit",99,2);
    DEF_SUIT(td,"leather armour",70,3);
    DEF_SUIT(td,"chain mail",90,4);
    DEF_SUIT(td,"plate mail",85,6);
    DEF_SUIT(td,"kevlar body armour",55,8);

    DEF_SHIELD(td,"small shield",30,4);
    DEF_SHIELD(td,"large shield",70,6);
    DEF_SHIELD(td,"riot shield",80,5);

    DEF_HEADGEAR(td,"paper hat",30,1);
    DEF_HEADGEAR(td,"paper crown",25,1);
    DEF_HEADGEAR(td,"flat cap",30,2);
    DEF_HEADGEAR(td,"beret",30,2);
    DEF_HEADGEAR(td,"tam-o'-shanter",30,2);
    DEF_HEADGEAR(td,"wooly hat",40,2);
    DEF_HEADGEAR(td,"bobble hat",40,2);
    DEF_HEADGEAR(td,"leather cap",30,3);
    DEF_HEADGEAR(td,"peaked cap",35,2);
    DEF_HEADGEAR(td,"skull cap",20,1);
    DEF_HEADGEAR(td,"baseball cap",35,1);
    DEF_HEADGEAR(td,"spiked helmet",45,6);
    DEF_HEADGEAR(td,"crash helmet",75,5);
    DEF_HEADGEAR(td,"hard hat",45,5);
    DEF_HEADGEAR(td,"mortarboard",30,1);
    DEF_HEADGEAR(td,"headscarf",40,1);
    DEF_HEADGEAR(td,"wig",45,1);
    DEF_HEADGEAR(td,"davy crocket hat",40,2);
    DEF_HEADGEAR(td,"deerstalker",50,2);
    DEF_HEADGEAR(td,"fez",38,2);
    DEF_HEADGEAR(td,"top hat",30,2);
    DEF_HEADGEAR(td,"trilby",32,1);
    DEF_HEADGEAR(td,"bowler hat",35,2);
    DEF_HEADGEAR(td,"steel helmet",85,6);
    DEF_HEADGEAR(td,"crown",40,4);

    DEF_GLOVES(td,"fingerless gloves",55,1);
    DEF_GLOVES(td,"mittens",90,2); /* Should reduce dex? Or evasion? */
    DEF_GLOVES(td,"gloves",85,1);
    DEF_GLOVES(td,"rigger's gloves",95,3);
    DEF_GLOVES(td,"kevlar gloves",95,7);
    DEF_GLOVES(td,"welder's gloves",100,4);/*should that be welder's gauntlets?*/
    DEF_GLOVES(td,"gauntlets",100,6);

    /* Good thing armour doesnt stack because neither do these names */
    DEF_FOOTWEAR(td,"socks",35,1);
    DEF_FOOTWEAR(td,"stockings",70,1);
    DEF_FOOTWEAR(td,"slippers",25,3);
    DEF_FOOTWEAR(td,"slipper-socks",30,2);
    DEF_FOOTWEAR(td,"plimsolls",25,3);
    DEF_FOOTWEAR(td,"sandals",15,2);
    DEF_FOOTWEAR(td,"clogs",25,4);
    DEF_FOOTWEAR(td,"trainers",30,3);
    DEF_FOOTWEAR(td,"shoes",30,3);
    DEF_FOOTWEAR(td,"low boots",32,4);
    DEF_FOOTWEAR(td,"high boots",40,4);
    DEF_FOOTWEAR(td,"cowboy boots",50,4);
    DEF_FOOTWEAR(td,"thigh boots",65,4);
    DEF_FOOTWEAR(td,"waders",90,3);
    DEF_FOOTWEAR(td,"wellies",50,5); /* Should reduce AGI or evasion */
    DEF_FOOTWEAR(td,"roller skates",30,2); /* Hmm, AGI also represents
    balance,etc , as well as speed. Roller skates should make you a bit
    clumsy but sometimes faster. Don't know how to do that */
    DEF_FOOTWEAR(td,"roller blades",35,2); /* Like roller skates */
    DEF_FOOTWEAR(td,"ice skates",35,2); /* Ditto, but only any use on ice */
    DEF_FOOTWEAR(td,"steel toecapped boots",40,6);

    DEF_RING("ring");
    DEF_RING("magic ring"); /* what more do I do for these??? */

    DEF_NECKWEAR(td,"necklace",0,0);
    DEF_NECKWEAR(td,"amulet",0,0);
    DEF_NECKWEAR(td,"torc",75,2);/* If neckwear is 2% covg, mebbe the
    coverage numbers should be out of 1000 instead to stop underflow */
    /* End of wieldables/wearables */

    /* Readables... Most of these are "flavoured" items */
    DEF_SCROLL("scroll"); /* er...*/

    DEF_BOOK("textbook");
    DEF_BOOK("novel"); /* just random stuff?*/
    DEF_BOOK("handbook"); /* gives some skills */
    DEF_BOOK("manual"); /* other skills*/
    DEF_BOOK("grimoire"); /* gives some spells */
    DEF_BOOK("codex"); /* other spells */
    /* End of readables */

    /* Comestibles!*/
    /* For most foods, thought must be given to how to number them */
    DEF_FOOD("sandwich");
    DEF_FOOD("scotch egg");
    DEF_FOOD("hard boiled egg");
    DEF_FOOD("pickle"); /* would this interfere with condiments?*/
    DEF_FOOD("fish"); /*should this be subdivided?*/
    DEF_FOOD("meat");
    DEF_FOOD("apple");
    DEF_FOOD("pancake");
    DEF_FOOD("orange");
    DEF_FOOD("doner kebab");
    DEF_FOOD("hamburger");
    DEF_FOOD("bread"); /* But *loaf* or *slice*? */
    DEF_FOOD("toast"); /* Must be slice, yes? */
    DEF_FOOD("banana");
    DEF_FOOD("shish kebab"); /* perhaps this could leave a skewer! */
    DEF_FOOD("noodles");
    DEF_FOOD("sushi");
    DEF_FOOD("meat pie");
    DEF_FOOD("quiche");
    DEF_FOOD("pizza");
    DEF_FOOD("pineapple");
    DEF_FOOD("soup"); /* things like these should have containers */
    DEF_FOOD("curry");
    DEF_FOOD("balti");
    DEF_FOOD("ice cream");
    DEF_FOOD("spam");
    DEF_FOOD("corned beef");
    DEF_FOOD("spaghetti");
    DEF_FOOD("rice");
    DEF_FOOD("packet of crisps");
    DEF_FOOD("banana chips");
    DEF_FOOD("bombay mix");
    DEF_FOOD("meatball");
    DEF_FOOD("haggis");
    DEF_FOOD("durian");
    DEF_FOOD("ham"); /* Specific meats */
    DEF_FOOD("gammon");
    DEF_FOOD("bacon");
    DEF_FOOD("beef");
    DEF_FOOD("steak");
    DEF_FOOD("lamb");
    DEF_FOOD("leg of lamb");
    DEF_FOOD("mutton");
    DEF_FOOD("pork chop");
    DEF_FOOD("lamb chop");
    DEF_FOOD("chicken breast");
    DEF_FOOD("turkey breast");
    DEF_FOOD("chicken leg");
    DEF_FOOD("turkey leg");
    DEF_FOOD("chicken wing");
    DEF_FOOD("turkey wing");
    DEF_FOOD("horse meat");  /* takes all sorts */
    DEF_FOOD("dog meat");  /* hmm */
    DEF_FOOD("dog food");
    DEF_FOOD("cat food");
    DEF_FOOD("soylent brown"); /* hehe... will have to include the
                                  story for this :) */

    /* Corpse types should be defined by monster definitions?
       Should corpses just be lumped in with food (which means normal food)?*/
    DEF_CONDIMENT("salt");
    DEF_CONDIMENT("black pepper");
    DEF_CONDIMENT("white pepper");
    DEF_CONDIMENT("szechuan pepper");
    DEF_CONDIMENT("cayenne pepper");
    DEF_CONDIMENT("malt vinegar");
    DEF_CONDIMENT("spirit vinegar");
    DEF_CONDIMENT("balsamic vinegar"); /* think that's all vinegars*/
    DEF_CONDIMENT("english mustard");
    DEF_CONDIMENT("french mustard");
    DEF_CONDIMENT("wholegrain mustard");
    DEF_CONDIMENT("dijon mustard"); /* Any other mustards I forgot?*/
    DEF_CONDIMENT("horseradish");
    DEF_CONDIMENT("japanese horseradish"); /* wasabi? */
    DEF_CONDIMENT("coriander");
    DEF_CONDIMENT("basil"); /* there are a few types of basil*/
    DEF_CONDIMENT("oregano");
    DEF_CONDIMENT("parsley");
    DEF_CONDIMENT("thyme");
    DEF_CONDIMENT("lemongrass");
    DEF_CONDIMENT("paprika");
    DEF_CONDIMENT("cinnamon");
    DEF_CONDIMENT("ginger");
    DEF_CONDIMENT("nutmeg");
    DEF_CONDIMENT("anise");
    DEF_CONDIMENT("cardamom");
    DEF_CONDIMENT("cloves"); /* do these really count as condiments? */
    DEF_CONDIMENT("garlic powder");
    DEF_CONDIMENT("powdered onion"); /* to distinguish from any food onion */
    DEF_CONDIMENT("grated cheddar");
    DEF_CONDIMENT("grated parmesan");
    DEF_CONDIMENT("sliced lemon peel");
    DEF_CONDIMENT("pesto"); /* yes I know theres lots of types */
    DEF_CONDIMENT("salsa");
    DEF_CONDIMENT("spaghetti sauce");
    DEF_CONDIMENT("tomato ketchup");
    DEF_CONDIMENT("brown sauce");
    DEF_CONDIMENT("HP sauce");
    DEF_CONDIMENT("soy sauce");
    DEF_CONDIMENT("worcestershire sauce");
    DEF_CONDIMENT("tabasco sauce");
    DEF_CONDIMENT("chilli sauce");
    DEF_CONDIMENT("lemon juice");
    DEF_CONDIMENT("lime juice");

    /* Some potions should start off known. The player would
    probably know only some. What he recognises would not be
    made clear to him. If the player starts off *not* recognising
    such a potion, that potion would take on a different name
    as given below. If the player DID recognise the potion from
    the beginning, those alternative names could be used by OTHER
    potions. You might only find out that a "clear potion" wasn't
    water when you find a "potion of water". */
    DEF_POTION("water"); /* may be "clear potion"*/
    DEF_POTION("orange juice"); /* may be "orange potion" */
    DEF_POTION("cola"); /* may be "black potion" */
    DEF_POTION("lemonade"); /* may be "fizzy potion" */
    DEF_POTION("booze"); /* may be "brown potion" */
    DEF_POTION("tea"); /* may be "brown potion"? */
    DEF_POTION("coffee"); /* may be "murky potion"? */

    DEF_POTION("purple potion"); /* Unknown types should be */
    DEF_POTION("green potion"); /* distinguished by a flag that */
    DEF_POTION("blue potion"); /* makes the word "potion" */
    DEF_POTION("yellow potion"); /* go either at the end, or at */
    DEF_POTION("red potion"); /* the beginning with the word "of" */
    DEF_POTION("turquoise potion");
    DEF_POTION("orange potion");
    DEF_POTION("brown potion");
    DEF_POTION("violet potion");
    DEF_POTION("pink potion");
    DEF_POTION("viscous potion");
    DEF_POTION("oily potion");
    DEF_POTION("fizzy potion");
    DEF_POTION("bubbling potion");
    DEF_POTION("fuming potion");
    DEF_POTION("cloudy potion");
    DEF_POTION("murky potion");
    DEF_POTION("clear potion");
    DEF_POTION("opaque potion");
    DEF_POTION("grey potion");
    DEF_POTION("black potion");
    DEF_POTION("sparkling potion");
    DEF_POTION("glowing potion");
    DEF_POTION("flashing potion");
    DEF_POTION("swirly potion");
    DEF_POTION("pulsing potion");

    /* Zappables... Again, flavoured items. */
    DEF_WAND("magic wand"); /* uh... */
    DEF_ROD("magic rod"); /* ... */
    /* end of zappables */

    /* The rest... */ 
    DEF_TOOL("key");
    DEF_TOOL("remote control"); /* does nothing! :) */
    DEF_TOOL("pen");
    DEF_TOOL("screwdriver"); /* repair stuff? untrap stuff? */
    DEF_TOOL("wire cutters"); /* untrapping stuff */
    DEF_TOOL("scissors"); /* shrug */
    DEF_TOOL("shovel");
    DEF_TOOL("bottle"); /* Can be filled to get a potion! Or condiment? */

    /* name, opens, wlim, iclass,extype */
    DEF_CONTAINER("bucket",0,100,IC_NOCLASSES,IG_STONE); /* for rocks! */
    DEF_CONTAINER("bag",0,30,IC_ALLCLASSES,IG_LASTGROUP);
    DEF_CONTAINER("paper bag",0,5,IC_COMESTIBLES,IG_COINAGE);
    DEF_CONTAINER("carrier bag",0,15,IC_ALLCLASSES,IG_LASTGROUP);
    DEF_CONTAINER("satchel",0,20,IC_READABLES,IG_FOOD);
    DEF_CONTAINER("rucksack",1,250,IC_ALLCLASSES,IG_LASTGROUP);
    DEF_CONTAINER("sack",0,75,IC_ALLCLASSES,IG_LASTGROUP);
    DEF_CONTAINER("quiver",0,20,IC_NOCLASSES,IG_MISSILE); /* for missiles */
    DEF_CONTAINER("wallet",1,10,IC_NOCLASSES,IG_COINAGE); /* for coinage */
    DEF_CONTAINER("purse",1,12,IC_VALUABLES,IG_LASTGROUP);  /* "  " */
    DEF_CONTAINER("money box",1,25,IC_NOCLASSES,IG_COINAGE); /* " " */
    DEF_CONTAINER("safe",1,70,IC_VALUABLES,IG_LASTGROUP); /* " " , heavy!*/
    DEF_CONTAINER("cardboard box",1,25,IC_ALLCLASSES,IG_LASTGROUP);
    DEF_CONTAINER("chest",1,400,IC_ALLCLASSES,IG_LASTGROUP); /* heavy! */
    DEF_CONTAINER("deposit box",1,15,IC_SMALL,IG_LASTGROUP); /* limited capacity! Not sure how to express that...*/
    DEF_CONTAINER("cruet",0,20,IC_NOCLASSES,IG_CONDIMENT); /* For condiments!!! */
    DEF_CONTAINER("mini fridge",1,40,IC_COMESTIBLES,IG_LASTGROUP); /* For food */
    DEF_CONTAINER("ice box",1,60,IC_COMESTIBLES,IG_LASTGROUP); /* " " , heavy! */
    DEF_CONTAINER("fridge freezer",1,200,IC_COMESTIBLES,IG_LASTGROUP); /* " " , V HEAVY! Hey, what about power?*/
    DEF_CONTAINER("chest freezer",1,400,IC_COMESTIBLES,IG_LASTGROUP); /* " ", V HEAVY! Ah, fudge it. */

    DEF_STONE("rock");
    DEF_STONE("pebble");
    DEF_STONE("flint");
    DEF_STONE("ore");
    DEF_STONE("gem");
    DEF_STONE("glass");

    DEF_COINAGE("gold coin");
    DEF_COINAGE("silver coin");
    DEF_COINAGE("brass coin");
    DEF_COINAGE("iron coin");
    DEF_COINAGE("thrupenny bit"); /* ;) */

}
