/* ==== ai_t.h ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Include mapgen_types.h before this. */

/* this header defines types used for the AI of Porrog's monsters.
   The AI basically works as a cache of data of which doors a
   monster has gone through, to speed up the monster's progress
   through the map. As such, it requires very very little processing,
   only data of which entrances lead where, etc, which gets produced
   when the map is made (since I changed the map generator to
   work in terms of regions and areas), but of course it also
   isn't effective in a map not built of rooms and tunnels
   (where some other AI function would need to be used instead).
   The scheme also doesn't really consitute an entire AI, only a
   means of getting each monster to cover as much of the dungeon
   as possible in a given amount of time, so they can find the
   player. Then they can do other stuff. */

/* Our lovely types */

typedef struct {
   int Xcoord,Ycoord; /* Where the door is. */
   int Age; /* Ticks since we last passed through this doorway */
   /* Or we could have a pointer to the entrance_t structure for
      the doorway.. But then it screws up when we change levels.
      The AI should know the cache elements aren't appropriate when
      they're on other levels, but when moving back to the level
      they're for, Porrog would be reconstructing the map from
      a description of the rooms and regions, and the pointers
      would be different. We could choose to have *both*
      representations, but I don't know what that buys us as yet. */
} aicache_doorent_t;

struct aicache_doorlist_t {
   aicache_doorent_t *Thisdoor;
   struct aicache_doorlist_t *Next;
};
typedef struct aicache_doorlist_t aicache_doorlist_t;

struct aicache_rooment_t {
   int Roomno; /* The room the cache is for */
   int Level; /* The map on which the cache entry is valid */
   int Age; /* Ticks since in the room */
   int Order; /* Number of doors in the room */
   /* Remember: The Age and Order of a room entry determines its value
      in the cache for the replacement algorithm! We prefer recent
      entries, and ones with more doors. */
   struct aicache_doorlist_t *Paths; /* Data for each way out of the room */
   struct aicache_rooment_t *Next; /* Pointer to next room entity in cache */
};
typedef struct aicache_rooment_t aicache_rooment_t;

struct aicache_tunnelent_t {
   region_t *Tunnel; /* The region of tunnels this cache is for.
      NB- I don't yet know how to preserve this data over level changes.
      Maybe specify coords of *a* point in the region, but this is only
      appropriate for rediscovering the region pointer. */
   int Level; /* The map this cache entry is valid on */
   int Age; /* Ticks since in the region */
   int Order; /* Number of doors in the region */
   /* Again, Age and Order of a tunnel entry determine its value
      in the same way they do for room entries */
   struct aicache_doorlist_t *Paths; /* Data for each way out of the tunnels */
   struct aicache_tunnelent_t *Next; /* Pointer to next tunnel entity in cache*/
};
typedef struct aicache_tunnelent_t aicache_tunnelent_t;

struct aicache_memory_t {
   aicache_rooment_t *Roomcache;
   aicache_tunnelent_t *Tunnelcache;
   /* Think this is all we'll have. */
   /* No, now adding singular buffer elements, to remove very stupid acts..*/
   aicache_rooment_t *Roombuffer;
   aicache_tunnelent_t *Tunnelbuffer;
   int Roombuf_flag; /* 1 if buffer current. 0 means stale. */
   int Tunnelbuf_flag;
};
typedef struct aicache_memory_t  aicache_memory_t;
