/* ==== dXrng.h ==== */
/* Declarations for an RNG for the roguelike game Porrog */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* We support the standard RPG dice:
   d4, d6, d8,  d10,d12,d20.
   Many roguelikes have d3 and d5, we will simulate these
   by rolling d6 or d10 and dividing. A lot of work? On the
   contrary, it means we create less tables... */


#define NUMSEQS 3
/* Describes how many lists of random number sequences we have for each
   type of dice. In the original proposal given in IDEA, this was 5, but
   I decided that would be iffy. */

struct roller {
    /* Listnum stays constant until we finish a sequence, then we get
       a new one. ?_Offset elements remain constant within a sequence too.*/
    byte ListNum; /*What section sequence list we're following */

    byte ListPtr; /* pointer into the list, to fetch a section number with*/
    byte S_Offset; /*Offset to modulate section number with */

    byte SecPtr; /* pointer into the section we look at, to get an entry*/
    byte E_Offset; /*Offset to modulate entry number with */

    byte R_Offset; /*Offset we modulate the returned roll with*/

    /* dX specific data, which stays constant within a roller. */
    byte dXType; /*What type of die we use: can be 3,4,5,6,8,10,12,20.*/
    byte N_Num; /*The number of sections in the table, which is also
                 the number of items in a section sequence list */
    byte *dXTab; /*Pointer to the roll table for this die type*/
    byte *SeqList; /*Pointer to the sequence lists for this die type */
};
/* Here, we've used the scheme of keeping the dXtype specific data within
  the roller structure, meaning the roller function can be totally generic.
  The main use of specifying the dXtype here is that we use it to modulus
  the roll given after applying the R_Offset. Now, the clever thing is,
  we can actually use this to convert, say a d6 to a d3, without adding
  any extra code!!!
  NB- using the modulus on rolls means we want to have rolls internally
  stored as range 0 to X-1 rather than 1 to X. */

typedef struct roller dXrng_roller;

extern dXrng_roller *d4set[];
extern dXrng_roller *d6set[];
extern dXrng_roller *d8set[];
extern dXrng_roller *d10set[];
extern dXrng_roller *d12set[];
extern dXrng_roller *d20set[];


void initialise_dXrng(); /* Populates the roll tables and sequence lists
                         for all dice types.  You want to call this
                         before using other dXrng functions. */

void make_dXen(); /* Fills the dXset[] arrays with new die structures
                      (using new_dX(). Should be called *after*
                      initialise_dXrng(), and before using any of those
                      dice. */

dXrng_roller *new_dX(int sides); /* Takes a number of sides for a new die,
                      returns a newly allocated and populated structure,
                      ready to use. Expects the roll-tables and sequence
                      lists to have already been set up! */

int dXrng_sysrandom(int range); /* The ultimate source of random numbers*/

int roll_dX(dXrng_roller *die); /* Rolls a die, returns the result (and
                      updates the pointers stored in the die roller
                      structure. Takes care of issues of picking new
                      sequence lists and offsets when a sequence is
                      finished. In fact, it takes care of practically
                      everything assuming the data has been  set up!*/
