/* ==== enums.h ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* Only has all the enumerated types. Depends on nothing else? */

typedef enum {
    PTL_PWALL, /* Permanent wall at the edge of the map*/
    PTL_ROCK, /* was PTL_TWALL*/
    PTL_HWALL, /* A horizontal perimeter wall (of a room)*/
    PTL_VWALL, /* " vertical    """""""*/
    PTL_CORNER,/* Corner of a room. Similar to PTL_PWALL.*/
    PTL_DOOR,
    PTL_ODOOR, /* A door that is open not closed. Doh.*/
    PTL_FLOOR,
    PTL_TUNNEL, /* Area of floor in a tunnel. Just for mapgen functions.*/
    PTL_STAIRSUP,
    PTL_STAIRSDOWN,
    PTL_ALTAR,
    PTL_RMID
} tile_t;

typedef enum {
   EQUIP_WEAPON, /* Your wieldables */
   EQUIP_RANGED,

   EQUIP_SUIT, /* Your armour */
   EQUIP_SHIELD,
   EQUIP_HEADGEAR,
   EQUIP_GLOVES,
   EQUIP_FOOTWEAR,

   EQUIP_RING1, /* Your jewellery */
   EQUIP_RING2, /* Might also have 2 other rings? */
   EQUIP_NECKWEAR,
   EQUIP_NONE /* Not equippable */
}equipspace_t; /* Position that an item of equipment is at */

typedef enum {
   IG_WEAPON, /* Wieldables first */
   IG_RANGED,
   IG_MISSILE,

   IG_SUIT,  /* Wearables */
   IG_SHIELD,
   IG_HEADGEAR,
   IG_GLOVES,
   IG_FOOTWEAR,

   IG_RING, /*Including jewellery */
   IG_NECKWEAR,

   IG_SCROLL, /* Readables */
   IG_BOOK, /* spellbooks and mundane books together or not?? */

   IG_FOOD, /* Comestibles */
   IG_CORPSE,
   IG_CONDIMENT, /* Can be applied to other foods, thrown, etc */
   IG_POTION,

   IG_WAND, /* Zappables */
   IG_ROD,

   IG_TOOL, /* Generic tools */
   IG_CONTAINER,
   IG_STONE, /* Gemstones and rocks and pebbles and so on */
   IG_COINAGE, /* That everything??*/
   IG_LASTGROUP /* Marker! */
} itemgroup_t;

typedef enum {
   IC_WIELDABLES,
   IC_WEARABLES,
   IC_WEAPONS,
   IC_ARMOURS,
   IC_JEWELLERY,
   IC_READABLES,
   IC_COMESTIBLES,
   IC_ZAPPABLES,
   IC_VALUABLES,
   IC_SMALL, /* Should perhaps be done differently??? */
   IC_ALLCLASSES,
   IC_NOCLASSES
} itemclass_t; /* Used for, eg, stating what items a container can carry */

typedef enum {
   FACE_ERROR,
   FACE_NONE,
   FACE_N,
   FACE_NE,
   FACE_E,
   FACE_SE,
   FACE_S,
   FACE_SW,
   FACE_W,
   FACE_NW,
   FACE_UP,
   FACE_DOWN,
   FACE_HERE
} direction_t;

typedef enum {
   PGOD_ENPITSU,
   PGOD_EB,
   PGOD_GRUBBLE,
   PGOD_URANUS,
   PGOD_PUTPUT,
   PGOD_NAPKIN,
   PGOD_KECKS,
   PGOD_SWUUJ,
   PGOD_K,
   PGOD_TINSEL,
   PGOD_HAPPYDUDE, /* Working title... */
   PGOD_TVB,
   PGOD_ATHEIST, /* Specifically when one *denies* gods */
   PGOD_SECULAR /* Used for when one hasn't chosen yet? */
} deity_t; /* The pantheon of Porrog! See NOTES/IDEA_DEITIES*/

typedef enum {
   MEM_LVLMAP,
   MEM_LVLSTRCT,
   /* low-level structures inside level */
   MEM_AREA,
   MEM_REGION,
   MEM_TUNNEL,
   MEM_INTERSECTION,
   MEM_ENTLIST,
   MEM_ENTRANCE,
   /* Items+mobs */
   MEM_ITYPE,
   MEM_ITEM,
   MEM_MTYPE,
   MEM_MONSTER,
   /* AI stuff */
   MEM_MMEM,
   MEM_TUNMEM,
   MEM_ROOMMEM,
   MEM_DOORMEM,
   MEM_TSEARCH,
   MEM_ISEARCH,
   /* Obscure stuff */
   MEM_LOCID,
   /* More generic stuff */
   MEM_STRING,
   MEM_LASTTYPE /* Marks the end */
} strgrp_t; /* Used with memutils functions to state what sort of elements
               are being mallocced */
typedef enum {
   LOGLEVEL_COMMENTARY, /* Pretty trivial */
   LOGLEVEL_DEBUG,   /* Not necessarily *bad*, but could be consequential */
   LOGLEVEL_WARNING, /* This really shouldn't happen, but hmm */
   LOGLEVEL_FOOBAR  /* So serious that before bugger() was written, such
                       an error would call crash() and end the game. */
} buglevel_t;

