/* ==== mg_utils.h ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/* Prototypes for map-building subfunctions that used to be in
   maparray.c, and are only used by the map-building functions.*/
/* This header used to contain the definitions for structures used
   in maparray.c, but is now just function prototypes. */


/* structures for map generation will be needed by the AI routines
  too, for pathfinding. They'll be in a separate typedefs header. */

/* We will ONLY have private functions here. */

extern roomdata_t room[];

void merge_areas(area_t *area1,area_t *area2);
/* MUST REMEMBER that this has to CORRECTLY update the area lists!!!*/

/* NEW FUNCTION for region modification */
void merge_regions(region_t *region1, region_t *region2);
/* Discussed in note Tunnels2: Currently I think this requires that
 both regions have the same area. If they don't, you should call
 merge_areas() FIRST. */

/* May or may not be useful to non-mapgen functions */
/* useful to anything else??? */
int pick_room(int totalrooms);/* tries to pick a room at random (using
   various methods) that is an actual created room. If it continually
   fails to, it just picks the earliest one in the list*/

int pick_room_from_area(area_t *area); /*As above, but only rooms from
some given area*/

/* attempt_joining is now declared in ajoin.h?*/


int probe_path(int x,int y, direction_t direction);
/* Starts at x,y, moves in direction. When it hits something that isn't
   PTL_ROCK, it returns tha point at which it finds it.
   If direction is east or west, the return value is an X coord.
   otherwise, the return value is a Y coord. */

void remake_regions(region_t *region);
/* Goes through all regions in a list structure, and remakes
  all its tunnels (carving them out on the grid, and placing
  the tunnel pointers), and places entrances. */

void remake_tunnels(tunnellist_t *tunnel);
/* Goes through all the tunnels in a region. Called by remake_regions*/

void remake_doors(entrancelist_t *door);
/* Goes throught all the doors in a region. Called by remake_regions */

void build_tunnel(region_t *region,int xstart,int ystart,int xend,int yend);
/* Tunnels out an area between start and end. Doesn't check where it's
  going (you should have done that already!). Then creates a tunnellist_t
  structure, and sticks it into area. DOESNT check whether to join areas!!
  Nope, NOW, sticks it into region, which is expected to have been created
  and to have area element filled in.*/

/* find_room() is useful outside map generation, so is in another header.*/

void replace_items(item_t *item);
/* Replaces items onto grid. Currently has funky cache system
   to speed up the algorithm so items can be placed in the
   proper order without being too slow! */

void add_room_items(int roomno);
/* Adds some item to room roomno */

void replace_mobs(monster_t *mob);
/* Replaces monsters onto grid */

void add_room_mobs(int roomno);
/* Adds some mob to room roomno */

void remake_features(levelstructs_t *level);
/* Remakes the features found in level */

int make_feature(int roomno,tile_t tile);
/* Places a feature tile in room roomno.
   Yes, that means they can't go in tunnels.
   Need a way to specify the extra data.
   Now, returns the index used for feature number*/

/* area_t *find_area() is changing... */
/* Should be changed to find_tunnel();? */
/* Almost. Now changed to find_region(). */
region_t *find_region(int x,int y);
 /*Takes a map coord, works out what it is, then
  if a room, scans all rooms to find it, else if a tunnel, scans all
  tunnel sections of all areas to find it!!! Probably slow.
  NOPE: Now, if it's a tunnel (or an entrance????) checks the tunnel-
  -pointer grid (runs parallel to the map grid) and finds ptr->Region,
  and returns it. If it's not (including if it's a room), returns NULL.
  Should now be very very quick.*/

entrance_t *find_door(region_t *region,int x,int y);
/* Finds door in a given region at given coords. */

entrance_t *find_common_door(region_t *region,int roomno);
/* If there is a door in region into roomno, return its pointer.
   Else return NULL */

area_t *choose_small_area(); /*Searches through whole area list, to find
  smallest area. If finds an area of size 1, is likely to just accept it,
  else will give minimum at end. */

void free_entrancelist(entrancelist_t *elist,int free_entrance);
/* Frees entrance list, and if free_entrance is set, its entrances too */
void free_areas(); /* Frees firstarea, and all its tunnels.
 When called, there should be no *other* areas, but if there are,
 removes them too. */

