/* ==== item.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include "C-foogrid.h"
#include "porrog_defs.h"
#include "memutils.h"
/* #include "maparray.h" Don't think we use this one any more.*/
#include "item_t.h"
#include "structures.h" /*Hers's where that stuff is now*/
#include "mapgen_types.h" /* Needed for levelmap type */
#include "maparray.h" /* Declares map */
#include "player_t.h"
#include "dXrng.h"/* for p_inv_init... should it be done within player.c? */

/* protos */
#include "item_p.h"
#include "player_p.h"
#include "crash.h"
#include "ui.h"

/* itemtype_t *first_it;  Pointer to the first item-type. N/A? */

/*itemtype_t *first_its[IG_LASTGROUP]; Now in globals.c */
/*int it_nums[IG_LASTGROUP]; Now in globals.c */

/* We now need some structure that keeps track of which inventory
   letters have and haven't been used. 52 possible inventory letters
   means we could have an array of 13 elements each describing
   ranges of letters: 4 each. When you want to get a new inventory
   letter, you search through the array from the start till you
   find one element where it's indicated that *at least* one letter
   is available.
   The array doesn't *necessarily* have to indicate *which* of those
   4 letters are available, as simply knowing that there are letters
   within some range allows you to search through the letters of the
   inventory and have some clue of which ones to look for.*/

/* Ah. This search, whilst more efficient than a blind search, is
   actually a bit involved. It's actually a lot easier to just
   have a bit-array... As this can be done with just 52 bits, it's
   hardly memory intensive! It could be done with 8 chars, but
   I'm using 13 to avoid having one half-filled. */

/* char invletter_array[13]; Now in globals.c */
extern char invletter_array[]; /* This's the only file to use this, but...*/

void p_init_inv()
{
   int counter;
   player.Inventlist=NULL;
   for (counter=0;counter<13;counter++) invletter_array[counter]=15;
   /* All free as unused */
   /* Now create some initial inventory items?? */
   switch(roll_dX(d4set[0])) {
      case 0:
        p_get_item(new_item(get_itemtype_by_name("pinstripe suit",IG_SUIT)));
        p_get_item(new_item(get_itemtype_by_name("bowler hat",IG_HEADGEAR)));
        p_get_item(new_item(get_itemtype_by_name("shoes",IG_FOOTWEAR)));
        p_get_item(new_item(get_itemtype_by_name("tea",IG_POTION)));
        p_get_item(new_item(get_itemtype_by_name("pen",IG_TOOL)));
        p_get_item(new_item(get_itemtype_by_name("gold coin",IG_COINAGE)));
        /* should have 4 or 5? Just as spare change... */
        break;
      case 1:
        p_get_item(new_item(get_itemtype_by_name("gimp suit",IG_SUIT)));
        p_get_item(new_item(get_itemtype_by_name("waders",IG_FOOTWEAR)));
        p_get_item(new_item(get_itemtype_by_name("whip",IG_WEAPON)));
        p_get_item(new_item(get_itemtype_by_name("soylent brown",IG_FOOD)));
        p_get_item(new_item(get_itemtype_by_name("paper bag",IG_CONTAINER)));
        p_get_item(new_item(get_itemtype_by_name("thrupenny bit",IG_COINAGE)));
        break;
      case 2:
        p_get_item(new_item(get_itemtype_by_name("sword",IG_WEAPON)));
        p_get_item(new_item(get_itemtype_by_name("leather armour",IG_SUIT)));
        p_get_item(new_item(get_itemtype_by_name("leg of lamb",IG_FOOD)));
        p_get_item(new_item(get_itemtype_by_name("brass coin",IG_COINAGE)));
        /* Should have quite a few? 50? */
        break;
      case 3:
        p_get_item(new_item(get_itemtype_by_name("penknife",IG_WEAPON)));
        p_get_item(new_item(get_itemtype_by_name("paper crown",IG_HEADGEAR)));
        p_get_item(new_item(get_itemtype_by_name("fingerless gloves",IG_GLOVES)));
        p_get_item(new_item(get_itemtype_by_name("stockings",IG_FOOTWEAR)));
        p_get_item(new_item(get_itemtype_by_name("necklace",IG_NECKWEAR)));
        p_get_item(new_item(get_itemtype_by_name("doner kebab",IG_FOOD)));
        p_get_item(new_item(get_itemtype_by_name("booze",IG_POTION)));
        p_get_item(new_item(get_itemtype_by_name("carrier bag",IG_CONTAINER)));
        break;
   }
}

char first_unused_invletter()
{
   int counter;
   char bitmap;
   int letter;
   for(counter=0;counter<13;counter++)
   {
      bitmap=invletter_array[counter];
      if (bitmap) break;
   }
   if (13==counter) {/* There wasn't one!!*/ return(' ');}
   letter=counter*4;
   for(;bitmap<8;letter++,bitmap*=2); /* Only the 4 LSB bits used.
    It's fairly important the 4 MBS bits are clear, but if they
    aren't, the first letter will appear to be free, which would
    initially be true... When stating that a letter is being used,
    the top bits get cleared in the process.*/
   /* We now have a letter number */
   if (letter<26)
   { /* Lower case letter */
      return('a'+letter);
   } else { /* Upper case */
      return('A'+(letter-26));
   }
}
void tag_invletter_usage(char letter,int discarding)
{ /* discarding   means you're freeing that letter.*/
   int index;
   int exponent;
   int normalised;
   if(letter>='a')
   {
      normalised=letter-'a';
   } else {
      normalised=(letter-'A')+26;
   }
   index=(normalised/4);
   exponent=normalised-(index*4);
   if (discarding)
   {   /* We're freeing this letter, set its bit */
      invletter_array[index] |= (8 >> exponent);
   } else { /* We're using this letter, so clear its bit */
      invletter_array[index] &= (15-(8 >> exponent));
   }
}

/* creates new item and returns pointer to it */
item_t *new_item(itemtype_t *otype)
{
   item_t *newitem;
   if(otype==NULL) {
     bugger(REPTYPE_ITEMS,LOGLEVEL_FOOBAR,"Tried to create new item with NULL type");
     crash(INCONSISTENT);
   }
   newitem=(item_t *)counting_malloc(MEM_ITEM,sizeof(item_t));
   newitem->Itemtype=otype;
   newitem->Invletter=' ';
   /* newitem->Next=NULL; -AFAICT this will get set when needed */
   /* newitem->Next2=NULL; -Ditto */
   return(newitem);
}

/* useful because of containers. DO NOT use on phonyitem! */
void free_item(item_t *item)
{
   item_t *contents;
   item_t *next;
   if(item->Itemtype->Type==IG_CONTAINER) {
      for(contents=item->Instdata.Container.Contents;contents!=NULL;contents=next) {
         next=contents->Next;
         free_item(contents);
      }
   }
   counting_free(MEM_ITEM,item);
}

void add_item(levelstructs_t *level,int x, int y,itemtype_t *otype)
{
   item_t *newitem;
   long offset=FOOGRID_IDX(map,y,x);
   maploc_t *mapstart=map->Data; /*Got access to coords*/
   /* newitem=(item_t *)counting_malloc(MEM_ITEM,sizeof(item_t));
   newitem->Itemtype=otype;
   newitem->Invletter=' '; No letter yet   now use new_item(). */

   newitem=new_item(otype);
   newitem->Next=(mapstart+offset)->Itemlist;
   (mapstart+offset)->Itemlist=newitem;
   newitem->X=x; newitem->Y=y; /* Remember where it is */
   /* Now, we put the item in the level-wide item list */
   newitem->Next2=level->Itemlist;
   level->Itemlist=newitem;
   /*Its on top of the stack there. We don't return anything */
}
/* Places a new instance of otype at the position given. Could
 this now use place_item()? */

void new_itemtype(itemtype_t *space,char symbol,char *name,itemgroup_t type)
{
    space->Symbol=symbol;
    space->Name=name; /*You'd better not free name later...*/
    space->Type=type;
    space->Weight=1; /* Set them for real later. */
    /* Hmm, is that all?*/
}

itemtype_t *push_itype(itemgroup_t group,char *name)
{
    itemtype_t *newit;
    char symbol;
    newit=(itemtype_t *)counting_malloc(MEM_ITYPE,sizeof(itemtype_t));
    /* newit->Next=first_it;  Changed to: */
    newit->Next=first_its[group];
    /* first_it=newit; Changed to: */
    first_its[group]=newit;
    it_nums[group]++; /* Update the count for this type */
    switch(group) {
       case IG_WEAPON:
         symbol=')';
         break;
       case IG_RANGED:
         symbol='}';
         break;
       case IG_MISSILE:
         symbol='-';
         break;
       case IG_SUIT:
         symbol='[';
/*       newit->Typedata.Arm.Cov=40;
         newit->Typedata.Arm.Prot=1; */
         break;
       case IG_SHIELD:
         symbol='(';
/*       newit->Typedata.Arm.Cov=15;
         newit->Typedata.Arm.Prot=3; */
         break;
       case IG_HEADGEAR:
         symbol='{';
/*       newit->Typedata.Arm.Cov=15; */
         break;
       case IG_GLOVES:
         symbol='{';
/*       newit->Typedata.Arm.Cov=8; */
         break;
       case IG_FOOTWEAR:
         symbol='{';
/*       newit->Typedata.Arm.Cov=20; */
         break;
       case IG_RING:
         symbol='=';
         break;
       case IG_NECKWEAR:
         symbol='='; /* As other jewelery?! */
/*       newit->Typedata.Arm.Cov=2;  */
         break;
       case IG_SCROLL:
         symbol='?';
         break;
       case IG_BOOK:
         symbol='+';
         break;
       case IG_FOOD:
         symbol='%';
         break;
       case IG_CORPSE:
         symbol='%';
         break;
       case IG_CONDIMENT:
         symbol='"'; /* Check. */
         break;
       case IG_POTION:
         symbol='!';
         break;
       case IG_WAND:
         symbol='/';
         break;
       case IG_ROD:
         symbol='|'; /*if we get staves, rods will be / too. */
         break;
       case IG_TOOL:
         symbol='&'; /* \ is another alternative...*/
         break;
       case IG_CONTAINER:
         symbol=']'; /* Could also be as tools. */
         break;
       case IG_STONE:
         symbol='*';
         break;
       case IG_COINAGE:
         symbol='$';
         break;
       case IG_LASTGROUP:
         bugger(REPTYPE_ITEMS,LOGLEVEL_FOOBAR,"push_itype():Item put in meaningless group!");
         crash(OUT_OF_BOUNDS);
         break;
    /* default: If we add new types, we should get
       a warning if we leave out a default */
    }
    new_itemtype(newit,symbol,name, group);
    return(newit);
}

/* It shouldn't be slow enough to need optimising, but I wouldn't
want to be using it too often...  The expected use is when setting
up the players inventory, or creating monsters. Or possibly wishes???
But this can't (yet) deal with plurals, etc. */
itemtype_t *get_itemtype_by_name(char *name,itemgroup_t grp)
{
    itemtype_t *scan;
    if (grp!=IG_LASTGROUP) {
      for(scan=first_its[grp];scan!=NULL;scan=scan->Next) {
         if(strcmp(scan->Name,name)==0) {
           bugger(REPTYPE_ITEMS,LOGLEVEL_COMMENTARY,"get_itemtype_by_name found item OK");
           return(scan);
         }
      }
    } else { /* search them all, return the first */
      for(grp=IG_WEAPON;grp<IG_LASTGROUP;grp++) {
         for(scan=first_its[grp];scan!=NULL;scan=scan->Next) {
            if(strcmp(scan->Name,name)==0) {
              bugger(REPTYPE_ITEMS,LOGLEVEL_COMMENTARY,"get_itemtype_by_name found item OK");
              return(scan);
            }
         }
      }
    }
    bugger(REPTYPE_ITEMS,LOGLEVEL_DEBUG,"get_itemtype_by_name couldn't find named type");
    return(NULL);
}

/* Removes top item from a heap on the floor, returns it (with
   that item individualised)*/
/* CHANGING to remove a SPECIFIC item from the heap, returning
   nothing. 
item_t *skim_item(foogrid *map,int x,int y) */
void skim_item(levelstructs_t *level, int x,int y, item_t *item)
{
   long offset=FOOGRID_IDX(map,y,x);
   maploc_t *mapstart=map->Data;
   item_t *heap;
   item_t *next; /* Next2, the next in level-wide itemlist */
   item_t *prev; /* item before item in heap */
   if(NULL==item)
   { /* Silly thing to do */
      crash(INCONSISTENT);
   }
   heap=(mapstart+offset)->Itemlist;
   if (NULL==heap)
   {
      /* skim_item now expects to remove a given item from
      the heap. If there's no heap, the item cant be there!!!*/
      crash(INCONSISTENT);
   }
   /* Old behaviour: Take away the TOP item.
   (mapstart+offset)->Itemlist=item->Next;
   item->Next=NULL;
   return(item);
   */
   if (heap==item)
   { /* Special case */
      (mapstart+offset)->Itemlist = item->Next;
   } else { /* item must be somewhere else */
      for(prev=heap; prev->Next!=item&&prev->Next!=NULL; prev=prev->Next);
      if(prev->Next==NULL) crash(INCONSISTENT); /* Not found!! */
      prev->Next = item->Next;
   }
   /* item is no longer on the floor. Hurrah! Caller MUST have either
      put it somewhere else (eg inventory) or freed it. */
   /* NOW also take it off the level-wide item list! */
   next=item->Next2;
   if (level->Itemlist==item) {
      level->Itemlist=next;
   } else {
      for(prev=level->Itemlist;prev->Next2!=item&&prev->Next2!=NULL;prev=prev->Next2);
      if(prev->Next2==NULL) crash(INCONSISTENT); /*Not found! */
      prev->Next2=next;
   }
}

/* Puts an item on top of heap on floor. Assumes item is alone.*/
void place_item(levelstructs_t *level,int x,int y,item_t *item)
{
   long offset=FOOGRID_IDX(map,y,x);
   maploc_t *mapstart=map->Data;
   if (NULL==item) return; /* To be safe*/
   item->Next=(mapstart+offset)->Itemlist;
   (mapstart+offset)->Itemlist=item;
   item->X=x; item->Y=y; /* Remember where it is */
   /* Now put the item back into the level-wide item list */
   item->Next2=level->Itemlist;
   level->Itemlist=item;
}

int isinclass(itemtype_t *type,itemclass_t cl)
{
   itemgroup_t gr=type->Type;
   switch (cl) {
      case IC_WIELDABLES:
        return(gr==IG_WEAPON||gr==IG_RANGED); /* I think this included others*/
      case IC_WEARABLES:
        return(gr==IG_SUIT||gr==IG_SHIELD||gr==IG_HEADGEAR||gr==IG_GLOVES||gr==IG_FOOTWEAR||gr==IG_RING||gr==IG_NECKWEAR);
      case IC_WEAPONS:
        return(gr==IG_WEAPON); /* Um... why? */
      case IC_ARMOURS:
        return(gr==IG_SUIT||gr==IG_HEADGEAR||gr==IG_GLOVES||gr==IG_FOOTWEAR);
      case IC_JEWELLERY:
        return(gr==IG_RING||gr==IG_NECKWEAR);
      case IC_READABLES:
        return(gr==IG_SCROLL||gr==IG_BOOK);
      case IC_COMESTIBLES:
        return(gr==IG_FOOD||gr==IG_CORPSE||gr==IG_CONDIMENT||gr==IG_POTION);
      case IC_ZAPPABLES:
        return(gr==IG_WAND||gr==IG_ROD);
      case IC_VALUABLES:
        return(gr==IG_STONE||gr==IG_COINAGE); /* not all stones though! */
      case IC_SMALL:
        return(gr==IG_COINAGE||gr==IG_RING||gr==IG_NECKWEAR); /* Others too*/
      case IC_ALLCLASSES:
        return(1);
      case IC_NOCLASSES:
        return(0);
   }
   bugger(REPTYPE_ITEMS|REPTYPE_CODE,LOGLEVEL_WARNING,"isinclass() screwed up somehow!");
   return(0);
}

/* --- end of item.c --- */

