/* ==== main.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "porrog_defs.h"
#ifdef RANDOMGAME
#include <stdlib.h>
#include <time.h>
#endif
#include <ncurses.h>
#include <unistd.h>
#include <stddef.h>
#include "C-foogrid.h" /* Used for foogrid*/

/*#include "mapgen_types.h" */
#include "maparray.h"
#include "crash.h"
#include "dXrng.h"
#include "memutils.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "player_t.h"

/* protos */
#include "makemap_p.h" /* Map generator. Was in maparray.h */
#include "item_p.h"
#include "player_p.h"
#include "monster.h"
#include "version.h"
#include "ui.h"
#include "apply.h"


int ctrl_loop(); /* Is in ctrl.c, but not worth making a header for ;) */


int main(int argc, char **argv)
{
    /*
    int x=(MAPWID/2)-25,y=(MAPHGT/2)-10,quit=0;
    char k; */
#ifdef RANDOMGAME
    unsigned int ts;
#endif
    int exit_status;
    watchflags=REPTYPE_ITEMS; /* Extra logs for map generation */
/*  levelmap_t *firstlevel; Now in globals.c */

    /* Set up curses stuff */
    if (initscr()==NULL)
    { /* crash(SCREEN_ERROR); crash() now goes through curses! */
      fprintf(stderr,"Yarg! Porrog couldn't initialise the terminal display! Quitting.\n");
      exit(1);
    }
    /* LINES should now store the number of lines in the
    current terminal. In wterm and the Linux console (AFAIK) this should
    be 25, but in xterm this is 24! */
    notimeout(stdscr, 1);
    noecho();
    cbreak();
    memutils_init(); /* initialises list of allocation counts */

#ifdef RANDOMGAME
    ts=(unsigned int) time(NULL);
    srand(ts);
#endif
    /* Set up our stuff */
    initialise_dXrng(); /* done *after* curses starts as it may use crash()*/
    make_dXen(); /* Creates our global arrays of dice */
    i_init(); /* make_map places objects */
    m_init();

    /* The first map. */
/*  firstlevel=(levelmap_t *)counting_malloc(sizeof(levelmap_t)); */
    firstlevel=level_skel();
    thislevel=firstlevel;
    tls=thislevel->Unpacked; /*Saves much typing and dereferencing...*/
    init_player(thislevel,(MAPHGT/2),(MAPWID/2));
    make_dmap(firstlevel,1); /* 1 for came down the stairs */


    p_init_inv();

    show_messages("Welcome to Porrog!");

    exit_status=ctrl_loop();
    /* Now does pretty much everything */

    endwin();
    printf("Exiting.\n");/* Should say reason for exiting */
    /* ctrl() should do the freeing ATM */
    printf("%ld unfreed allocations made.\n",check_allocs());
    exit(0);
}
