/* ==== maparray.c ==== */
/* Now contains just the "public" functions, rest are in mapgen.c */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
/* #include <ncurses.h> Should be removed when debugging syms removed*/
#include "porrog_defs.h"
#include "C-foogrid.h"
#include "memutils.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h" /*We need this, but maybe just for the #defines*/
#include "player_t.h"

/* protos */
#include "makemap_p.h" /* Our protos! */
#include "mg_utils.h"
#include "ajoin.h"
#include "item_p.h"
#include "crash.h"
#include "dXrng.h"
#include "ui.h" /* For error messages, which really ought to
                   be in crash.h */

/*
foogrid *map;
foogrid *tmap;
  See all those? They're now in globals.c */

extern foogrid *tmap;
/* extern area_t *firstarea; replaced by tls->Firstarea; */

/* dXrng_roller *d12;
dXrng_roller *d20;  We're using the new dice arrays *now*. */

void mkmap_p1();
int mkmap_p2();
void mkmap_p3();
void mkmap_p4(int totalrooms);
void remake_rooms();
void remake_areas(area_t *area);

levelmap_t *level_skel()
{
   levelmap_t *this;
   levelstructs_t *that;
   this=(levelmap_t *)counting_malloc(MEM_LVLMAP,sizeof(levelmap_t));
   if(NULL==this) crash(NO_MEM); /*OW! */
   that=(levelstructs_t *)counting_malloc(MEM_LVLSTRCT,sizeof(levelstructs_t));
   if(NULL==that) crash(NO_MEM);
   this->Unpacked=that;
   this->Archived=0; /* Nice and fresh :) */
   this->Packed=NULL;
   this->Nextlevel=NULL; /* Or should caller set this? */
   /* I'll try this instead... */
     this->Nextlevel=firstlevel;
     firstlevel=this; /* Tag us onto the front of the level list */
   that->Nfeatures=0;
   that->Nmobs=0;
   that->Mobticker=0;
   that->Totalrooms=0;
   that->Itemlist=NULL;
   that->Moblist=NULL;
   /* Will prolly leave Totalrooms, etc to be set by make_dmap, etc */
   return(this);
}

void make_dmap(levelmap_t *this, int dir)
{
   /* Will change dir to be dir_t instead */

   int f,totalrooms; /* used here. */


   /* We just call the subparts in order.
      This was just split up so we could profile it all
      and find out where the astronomical slowdown appears
      (it was something like 90% of the whole time) */
   bugger(REPTYPE_MAPGEN,LOGLEVEL_COMMENTARY,"--Generating new map--");
   mkmap_p1();
   totalrooms=mkmap_p2();
   tls->Totalrooms=totalrooms; /* Just for future reference */
   mkmap_p3();
   /* We add the features HERE for simplicity */
   /* We'll change this so that we can get the coords of
   the stairs we came from, set the player's coords to
   them, and set the return path! */
   f=make_feature(216,dir?PTL_STAIRSUP:PTL_STAIRSDOWN);
   tls->Features[f].Extra.Destination->Map=lastlevel;
   tls->Features[f].Extra.Destination->X=player.X; /* Set the path back to*/
   tls->Features[f].Extra.Destination->Y=player.Y; /* Where you'd been */
   player.X=tls->Features[f].Xcoord; /*Put you on the new stairs */
   player.Y=tls->Features[f].Ycoord;
   make_feature(pick_room(tls->Totalrooms),dir?PTL_STAIRSDOWN:PTL_STAIRSUP);
   make_feature(pick_room(tls->Totalrooms),PTL_ALTAR);
   mkmap_p4(totalrooms);
#ifdef ONE_MON
   add_room_mobs(0); /* Just one */
#elif FEW_MON
   for (f=0;f<216;f++) if (tls->Room[f].Xcentre!=1 &&roll_dX(d6set[3])==0) add_room_mobs(f);
#else
   for (f=0;f<216;f++) if (tls->Room[f].Xcentre!=-1) add_room_mobs(f);
#endif
}

void remake_dmap(levelmap_t *this)
{
   /* Hey, do we even NEED an argument? */
   mkmap_p1(); /* Blank the grid. */
   /* Ignore mkmap_p2(), which decides where to put rooms */
   remake_rooms(); /* Replaces mkmap_p3() */
   remake_areas(tls->Firstarea); /* Go through each area in
      the list, and do what is needed to recreate it! */
   /* Done. */
   /* No, NOT quite. Still need mobs and items. */
   replace_mobs(tls->Moblist);
   replace_items(tls->Itemlist);
}

void mkmap_p1()
{
   int x,y;
   maploc_t *mapstart;
   maploc_t * elem;


   /* Shouldn't we only do this once??? */
   /* OF COURSE WE SHOULD!!! */
   if (NULL==map){
     map=create_foogrid(MAPHGT,MAPWID,sizeof(maploc_t));
     tmap=create_foogrid(MAPHGT,MAPWID,sizeof(tunnellist_t*)); /*Pointers to em*/
   }
   /* Now we populate that grid */
   mapstart=map->Data;

   for(x=0,elem=mapstart+FOOGRID_IDX(map,0,0); x<MAPWID; elem++,x++)
   {
      elem->Tiletype=PTL_PWALL; /* Can't be passed or mined */
      elem->Mptr=NULL;
      elem->Seen=0;
      elem->Itemlist=NULL;
      elem->Remembered=NULL; /* Doesn't seem to matter much, but...*/
/*    elem->Dest=NULL; Now we have a union instead. what we do then? */
   }
   for(x=0,elem=mapstart+FOOGRID_IDX(map,(MAPHGT-1),0); x<MAPWID; elem++,x++)
   {
      elem->Tiletype=PTL_PWALL;
      elem->Mptr=NULL;
      elem->Seen=0;
      elem->Itemlist=NULL;
      elem->Remembered=NULL;
/*    elem->Dest=NULL; Now we have a union instead. */
   }
   for(y=1;y<MAPHGT-1;y++)
   {
      for(x=0,elem=mapstart+FOOGRID_IDX(map,y,0);x<MAPWID;elem++,x++)
      {
         if(x==0||x==(MAPWID-1)) elem->Tiletype=PTL_PWALL; else elem->Tiletype=PTL_ROCK;
         elem->Mptr=NULL;
         elem->Seen=0;
         elem->Itemlist=NULL;
         elem->Remembered=NULL;
/*       elem->Dest=NULL; Now we have a union instead. */
      }
   }
}

/* Does phase 2 of make_map() routine */
int mkmap_p2()
{
   maploc_t *mapstart;
   maploc_t * elem;
   int n,s,x,y,dx,dy,flag, totalrooms;
   int tx,ty;/* Targets to scan to */

   mapstart=map->Data;
   /* Phase 2: Pick room centres (with enough space) */
   (mapstart+FOOGRID_IDX(map,MAPHGT/2,MAPWID/2))->Tiletype=PTL_RMID;
   tls->Room[216].Xcentre=MAPWID/2;
   tls->Room[216].Ycentre=MAPHGT/2;
   tls->Room[216].Firstentrance=NULL; /* Had forgotten this. */
   tls->Room[216].Area=(area_t *)counting_malloc(MEM_AREA,sizeof(area_t));
     /* Make sure there's a room in the middle (start position) */
   tls->Room[216].Area->Firstroom=216;
   /* NOW, tunnellists are in region data, and not areas. */
/* room[216].Area->Firsttunnel=NULL;  Now in region structures. */
   tls->Room[216].Area->Firstregion=NULL; /*But we do use this! */
   tls->Room[216].Area->Size=1;
   tls->Room[216].Area->Nextarea=NULL; /*End of area list. */
   tls->Firstarea=tls->Room[216].Area;
   totalrooms=1;

   /* Now we make a load more! */
   for(n=0;n<216;n++)
   {
      /* Pick some random spot, then check the squares +-6 squares around it
         for other spots that have already been marked. If we find any
         in *that* area, we don't use the spot we have, but look for another*/
      tls->Room[n].Xcentre=-1;tls->Room[n].Ycentre=-1; /* If we can't alloc one */
      tls->Room[n].Firstentrance=NULL;
      tls->Room[n].Area=NULL;
      for(s=14;s>7;s--)
      {
        /* was using 8+roll_dX(d20)++roll_dX(d8)*/
        /* changed 8 and d8 to 6 and d6, with shrink to 140X map.*/
        x=10+( (6*roll_dX(d20set[0]))+(roll_dX(d12set[0])%6) );
        y=10+( (6*roll_dX(d20set[0]))+(roll_dX(d12set[0])%6) );
        flag=0;
        tx=x+s; ty=y+s; /* Don't calculate them *each time*! */
        for(dy=y-s;dy<=ty;dy++)
        {
           if(flag==1) break; /* Don't scan the whole bloody thing!*/
           if(dy<0||dy>MAPHGT-1) continue;
           elem=mapstart+FOOGRID_IDX(map,dy,x-s);
             /* Could this go out of range? Would it matter? */
           for(dx=x-s;dx<=tx;dx++,elem++)
           {
              if(dx<0||dx>MAPWID-1) continue;
               /* this doesn't make a room here illegal*/
       /*     if( (mapstart+FOOGRID_IDX(map,dy,dx))->Tiletype!=PTL_ROCK) */
              if( (elem)->Tiletype!=PTL_ROCK)
              {
                 /* We can't build a room around here. */ flag=1;
                 break;
              }
           }
        }
        if (flag==0)
        {
           (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_RMID;
           tls->Room[n].Xcentre=x;tls->Room[n].Ycentre=y;
           tls->Room[n].Area=(area_t *)counting_malloc(MEM_AREA,sizeof(area_t));
           tls->Room[n].Area->Firstroom=n;
           /* AGAIN: Tunnellists are now just in regions, not areas*/
  /*       room[n].Area->Firsttunnel=NULL; Now in region structures */
           tls->Room[n].Area->Firstregion=NULL; /* But we have this now! */
           tls->Room[n].Area->Size=1;
     /*    room[n].Area->Nextarea=room[216].Area->Nextarea;
           room[216].Area->Nextarea=room[n].Area;*/
           tls->Room[n].Area->Nextarea=tls->Firstarea;
           tls->Firstarea=tls->Room[n].Area;
           /*Attach this area to head
             of area list. Well, OK, 216's area is the head of the list.
             So this goes *behind* the head. Meh.*/
           totalrooms++;
           break; /* Don't look for smaller spaces now. */
        }
      }
   }
   return(totalrooms);
}

void mkmap_p3()
{
   int n,s, x,y,  dx,dy, flag;
   int mr,mxr,myr;
   tile_t thistile;
   maploc_t *mapstart=map->Data;
   maploc_t * elem;


   /* Phase 3: */
   /* Room midpoints established. Find their dimensions!*/
   for(n=0;n<217;n++)
   {
      if (-1==tls->Room[n].Xcentre) continue;
      for(s=14;s>7;s--) /* s of 6 would mean no space between walls*/
      {
         flag=0;
         for(dy=-s;dy<=s;dy++)
         {
            y=tls->Room[n].Ycentre+dy;
            if(y<0||y>MAPHGT-1) continue;
            elem=(mapstart+FOOGRID_IDX(map,y,tls->Room[n].Xcentre-s));
            for(dx=-s;dx<=s;dx++,elem++)
            {
               x=tls->Room[n].Xcentre+dx;
               if(x<0||x>MAPWID-1||(dx==0&&dy==0)) continue;
           /*  if( (mapstart+FOOGRID_IDX(map,y,x))->Tiletype==PTL_RMID) */
               if( elem->Tiletype==PTL_RMID)
               {
                   flag=1; /*We've encountered the middle of another room*/
                   break;
               }
            }
            if(flag==1) break;
         }
         if (flag==0) break; /*We have a room up to size s that's clear*/
      }
      /* So we now know the maximum size, but we have to check for
         other excavations, to reduce the maximum x and y size */
      mr=(4*s)/7; /* max possible radius */
      /* if s-mr<4 mr=s-4 */
      mr=((s-mr<5)?s-5:mr);

      do {
        flag=0;
        myr=((roll_dX(d12set[0])*(mr-2))/17)+5; /* We say 5 to leave extra wall space*/
        /* rather than dividing by 11, we divide by 17 to give
           an aspect ratio that looks more square- in fact, more
           horizontal than square!*/
        mxr=((roll_dX(d12set[0])*(mr-2))/11)+5; /* We take 3 off if it's safe*/
        /*myr=(mrmax of 2/3 s, min of 2) */
        /* mxr=(max of 2/3 s, min of 2) */
        /* radius of 2 gives room size (internal) of 5,  wall size of 7*/
        /* Could be optimised too */
        for(dy=-myr;dy<=myr;dy++)
           for(dx=-mxr;dx<=mxr;dx++)
           {
              x=tls->Room[n].Xcentre+dx;y=tls->Room[n].Ycentre+dy;
              if(x<1||x>(MAPWID-2)||y<1||y>(MAPHGT-2)) {flag=1; break;}
              if(dx==0&&dy==0) continue; /*The middle of the room */
              thistile=(mapstart+FOOGRID_IDX(map,y,x))->Tiletype;
              if( thistile!=PTL_ROCK&&thistile!=PTL_HWALL&&thistile!=PTL_VWALL&&thistile!=PTL_CORNER)
              /* The perimeter walls *used* to be the same type as the rock*/
              {
                  flag=1;
                  break;
              }
           }
      }while(flag==1); /* we now have suitable dimensions! */
      myr=myr-3;mxr=mxr-3; /* Remove the wall compensation to draw it. */

      /*Proceed to dig out the room */
      /* This could be optimised too */
      for(dy=-myr;dy<=myr;dy++)
         for(dx=-mxr;dx<=mxr;dx++)
         {
            x=tls->Room[n].Xcentre+dx;y=tls->Room[n].Ycentre+dy;
            if((mapstart+FOOGRID_IDX(map,y,x))->Tiletype!=PTL_RMID) (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_FLOOR;
         }
      /* Hollowed out?*/
      tls->Room[n].Width=3+(mxr*2); /* mxr is horizontal radius, not width,
         and doesn't count the centre, nor the perimeter walls. */
      tls->Room[n].Height=3+(myr*2); /*Likewise*/
      tls->Room[n].Xstart=tls->Room[n].Xcentre-(mxr+1);
      tls->Room[n].Ystart=tls->Room[n].Ycentre-(myr+1);
      /* build the perimeter walls */
      dy=(myr+1);
      /* This could be optimised a little */
      for(dx=-mxr;dx<=mxr;dx++)
      {
         x=tls->Room[n].Xcentre+dx;y=tls->Room[n].Ycentre-dy;
         (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_HWALL;
         y=tls->Room[n].Ycentre+dy;
         (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_HWALL;
      }
      dx=(mxr+1);

      x=tls->Room[n].Xcentre-dx;y=tls->Room[n].Ycentre-dy;
      (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_CORNER;

      x=tls->Room[n].Xcentre+dx;
      (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_CORNER;

      y=tls->Room[n].Ycentre+dy;
      (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_CORNER;

      x=tls->Room[n].Xcentre-dx;
      (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_CORNER;

      /* Perhaps this could be optimised a tad? */
      for(dy=-myr;dy<=myr;dy++)
      {
         x=tls->Room[n].Xcentre-dx;y=tls->Room[n].Ycentre+dy;
         (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_VWALL;
         x=tls->Room[n].Xcentre+dx;
         (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_VWALL;
      }

      add_room_items(n);
/*    add_room_mobs(n); Now done after rooms are joined, to help aicache*/
      tls->Room[n].Nextroom=-1;
      /* Now we'd do stuff about links to other rooms. */
   }
}

void remake_rooms()
{
   /* Roughly adapted from mkmap_p3 */
   int n,x,y,sx,sy,tx,ty;
   maploc_t *mapstart=map->Data;
   for(n=0;n<217;n++)
   {
      if(-1==tls->Room[n].Xcentre) continue;
      /* OK, so excavate the bugger */
      sx=tls->Room[n].Xstart;
      sy=tls->Room[n].Ystart;
      tx=sx+tls->Room[n].Width-1;
      ty=sy+tls->Room[n].Height-1;
      for(y=sy;y<=ty;y++)
         for(x=sx;x<=tx;x++)
         {
            (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_FLOOR;
         }
      /* Not quite how mkmap_p3 does it... no comment ;) */
      /* Now do the walls. */
      for (x=sx;x<=tx;x++)
      {
         (mapstart+FOOGRID_IDX(map,sy,x))->Tiletype=PTL_HWALL;
         (mapstart+FOOGRID_IDX(map,ty,x))->Tiletype=PTL_HWALL;
      }
      (mapstart+FOOGRID_IDX(map,sy,sx))->Tiletype=PTL_CORNER;
      (mapstart+FOOGRID_IDX(map,sy,tx))->Tiletype=PTL_CORNER;
      (mapstart+FOOGRID_IDX(map,ty,sx))->Tiletype=PTL_CORNER;
      (mapstart+FOOGRID_IDX(map,ty,tx))->Tiletype=PTL_CORNER;
      for (y=sy;y<=ty;y++)
      {
         (mapstart+FOOGRID_IDX(map,y,sx))->Tiletype=PTL_VWALL;
         (mapstart+FOOGRID_IDX(map,y,tx))->Tiletype=PTL_VWALL;
      }
      /* I'm sure I did that pretty craply, but surely bettar
        than mkmap_p3 did! */
      /* We don't update the Room[] array at all, nor add items/mobs? */
      /* I think we do place the features here tho */
      remake_features(tls);
   }
}

/* Room for optimisation but doesn't seem to matter at all */
void mkmap_p4(int totalrooms)
{
   int done=0, n,tries,target;

/*   We don't really want this now, do we?
     move(1,0);addstr("     ");
     move(2,0);addstr("   "); */

   /* Phase 4: connect rooms so *everything* is accessible */
   while(done==0)
   {
       n=pick_room(totalrooms);
       for(tries=0;tries<5;tries++)
       {
          target=pick_room(totalrooms);
          if (target!=n&&(abs(tls->Room[n].Xcentre-tls->Room[target].Xcentre)<35)&&(abs(tls->Room[n].Ycentre-tls->Room[target].Ycentre)<35)&&(tls->Room[n].Area!=tls->Room[target].Area))
              /*EUREKA! We have one. Somehow. Indicate this fact*/
              tries=6; /*Do we still need to break here?*/
       }
       if(7==tries) /* perhaps this should be 7 if above is 6?? */
       {
          /* This means we have randomly found 2 rooms that
             are suitable for joining.*/
          if(tls->Room[n].Xcentre==-1) show_messages("Source room doesnt exist!");
          if(tls->Room[target].Xcentre==-1) show_messages("Target room doesnt exist!");
          attempt_joining(n,target); /*we won't *necessarily* do so!*/
          /* if path blocked by something not yet accessible from n,
            try to connect to that instead!*/
          /* else path must be blocked by something already accessible
          from n, so we try to steamroller past any tunnels we find
          to see if we can it a room accessible from n. Then we try
          to see if the path between that room and target it blocked.*/
       } else {
          /*We obviously failed to find a suitable target. Take
            a different approach. */
          /* find the smallest area. Then pick a room from that
          area that can make a tunnel to somewhere it cant yet
          access. Each time it hits a room already connected,
          restart the attempt from that room.*/
          n=pick_room_from_area(choose_small_area());
          for(tries=0;tries<15;tries++)/* Should be enough?*/
          {
             target=pick_room(totalrooms);
             if (target!=n && tls->Room[n].Area!=tls->Room[target].Area)
             /* We don't care where it is, if it's too far we'll
               connect to somewhere else anyway*/
                tries=19;
          }
          if (15==tries) /*We couldn't find one?? Search harder.*/
          {
             for(tries=0;tries<217;tries++)/*tries=roomno*/
             {
                if(tries!=n &&tls->Room[tries].Xcentre!=-1&&tls->Room[n].Area!=tls->Room[tries].Area)
                {
                   target=tries;
                   break;
                }
             }
             if(tls->Room[tries].Xcentre==-1)
             { show_messages("Picked nonexistent room!");crash(NO_MEM);}
             if (217==tries) crash(INCONSISTENT); /* We couldnt find one! */
          }
          if(target==-1) {show_messages("make_map:target=-1!");crash(NO_MEM);}
          attempt_joining(n,target);
       }

       /* If we find the first area we can find has a size of
         totalrooms, then everything is connected, so done=1;
         NB: room 216 is GUARANTEED to exist (room 0 is not!!!)
         so might as well test room[216].Area->Size for this.*/
/*      if ((room[216].Area->Size)>2) done++;
*/
       /* For the moment, we show which cues showed
          that everything was connected.
          small_area->Size>1 turned out to terminate map construction
          prematurely about one time in 20 or so, so I removed it. */
       if(tls->Firstarea->Size==totalrooms){
          /*move(1,0);addstr("first");*/done=1;}
       /* The first area contains all the rooms. */
       if(tls->Firstarea->Nextarea==NULL) {
          /*move(2,0);addstr("one");*/done=1;}
       /* The first area is the only area. */
       /* If at some point, some trigger and others dont, that's
         something to find out about.
         But "first" and "one" seem equivalent. */
   }
   /* We should now have everything fully accessible! */
/* free_areas();  Tidy up!*/
   /* Of course, with the new archiving scheme, we actually
   want to KEEP this data now.
   But we might convert it to bitstream format at some point.
   Now, our new "level_t" or "map_t" structures (whatever we
   call em) will have pointers and arrays for these to go in,
   and also for any bitstream. There'd be a flag to indicate
   whether the map is archived or not. */
}

void remake_areas(area_t *area)
{
   /* Mostly just a wrapper function. */
   for(;area!=NULL;area=area->Nextarea)
      remake_regions(area->Firstregion);
      /* remake_regions should act on ALL the regions in area's list */
   /* Er, that's it now? Seriously? */
}

