/* ==== mapgen.c ==== */
/* Adapted from maparray.c, contains the generator's helper functions */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
/* #include <ncurses.h> Should be removed when debugging syms removed*/
#include "porrog_defs.h"
#include "C-foogrid.h"
#include "memutils.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h" /*We need this, but maybe just for the #defines*/
#include "ai_t.h"

/* protos */
#include "ai_p.h"
#include "mg_utils.h"
#include "ajoin.h"
#include "item_p.h"
#include "monster.h"
#include "crash.h"
#include "dXrng.h"
#include "ui.h" /* For error messages, which really ought to
                   be in crash.h */

/*foogrid *map;
foogrid *tmap;
All now in globals.c */
extern foogrid  *tmap;
/* extern area_t   *firstarea; Replacing with tls->Firstarea */



int pick_room(int totalrooms)
{
    int roomno,count,seen;
    count=(totalrooms*((12*roll_dX(d20set[0]))+roll_dX(d12set[0]))/240);
    /* HOPEFULLY count should get truncated such that count < totalrooms.*/
    /* If not, we will likely get an out of bounds. Check, in case. */
    if (count>=totalrooms) show_messages("Picked room higher than totalrooms");
    for (roomno=-1,seen=0;seen<=count;roomno++,seen+=(tls->Room[roomno].Xcentre!=-1)?1:0);
    if (roomno>216||roomno==-1) show_messages("Picked impossible room");
    if (tls->Room[roomno].Xcentre==-1) show_messages("Picked unmade room");
    return(roomno);
}

int pick_room_from_area(area_t *area)
{
    /* Much the same as pick_room()*/
    int roomno,count,seen,totalrooms=area->Size;
    count=(totalrooms*((12*roll_dX(d20set[0]))+roll_dX(d12set[0]))/240);
    if (count>=totalrooms) crash(OUT_OF_BOUNDS);
    for (roomno=0,seen=0;seen<count;roomno=tls->Room[roomno].Nextroom)
        seen+=(tls->Room[roomno].Xcentre!=-1)?1:0;
    if (roomno>216||-1==roomno) crash(OUT_OF_BOUNDS);
    /* Presumably if we get a -1, it should crash anyway??*/
    if(tls->Room[roomno].Xcentre==-1) crash(OUT_OF_BOUNDS);
    return(roomno);
}

/* Finally cut out the old (and broken) version of attempt_joining() from here*/

int probe_path(int x,int y,direction_t direction)
{
    int crs;
    maploc_t *mapstart=map->Data;
    switch(direction)
    {
        case FACE_N:
          for(crs=y-1;crs>1;crs--)
             if ((mapstart+FOOGRID_IDX(map,crs,x))->Tiletype!=PTL_ROCK||(mapstart+FOOGRID_IDX(map,crs,x-1))->Tiletype==PTL_TUNNEL||(mapstart+FOOGRID_IDX(map,crs,x+1))->Tiletype==PTL_TUNNEL)
                break;
          if(crs==y)crash(NO_MEM);
          break;
        case FACE_E:
          for(crs=x+1;crs<MAPWID;crs++)
             if ((mapstart+FOOGRID_IDX(map,y,crs))->Tiletype!=PTL_ROCK||(mapstart+FOOGRID_IDX(map,y-1,crs))->Tiletype==PTL_TUNNEL||(mapstart+FOOGRID_IDX(map,y+1,crs))->Tiletype==PTL_TUNNEL)
                break;
          break;
        case FACE_S:
          for(crs=y+1;crs<MAPHGT;crs++)
             if ((mapstart+FOOGRID_IDX(map,crs,x))->Tiletype!=PTL_ROCK||(mapstart+FOOGRID_IDX(map,crs,x-1))->Tiletype==PTL_TUNNEL||(mapstart+FOOGRID_IDX(map,crs,x+1))->Tiletype==PTL_TUNNEL)
                break;
          break;
        case FACE_W:
          for(crs=x-1;crs>1;crs--)
             if ((mapstart+FOOGRID_IDX(map,y,crs))->Tiletype!=PTL_ROCK||(mapstart+FOOGRID_IDX(map,y-1,crs))->Tiletype==PTL_TUNNEL||(mapstart+FOOGRID_IDX(map,y+1,crs))->Tiletype==PTL_TUNNEL)
                break;
          break;
        default:
          crash(-1); /*Should find a better value!*/
    }
    return(crs);
}

void remake_regions(region_t *region)
{
    for (;region!=NULL;region=region->Nextregion)
    {
       /* First remake the tunnels */
       remake_tunnels(region->Firsttunnel);
       /* After ALL tunnels in the region, remake all entrances */
       remake_doors(region->Firstentrance);
    }
}

void remake_tunnels(tunnellist_t *tunnel)
{
    int x,y,tx,ty;
    maploc_t *mapstart=map->Data;
    tunnellist_t **tmapstart=tmap->Data;

    for (;tunnel!=NULL;tunnel=tunnel->Nexttunnel)
    {
       /* Now, each tunnel goes a bit like build_tunnel,
         except we READ details from the tunnellist_t
         structure rather than filling it in */
       x=tunnel->Xstart;
       tx=tunnel->Xend;
       if(x==tx){
          /* dir=FACE_S; */
          /* x=tunnel->Xstart; */
          ty=tunnel->Yend;
          for(y=tunnel->Ystart;y<=ty;y++)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=tunnel;
          }
       } else {
          /* dir=FACE_E; */
          y=tunnel->Ystart;
          /* tx=tunnel->Xend; */
          for (x=tunnel->Xstart;x<=tx;x++)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=tunnel;
          }
       }
       /* Well, that was less painful than I'd expected... */
    }
}

void remake_doors(entrancelist_t *door)
{
    entrance_t *e;
    maploc_t *mapstart=map->Data;
    for (;door!=NULL;door=door->Next)
    {
        e=door->Entrance;
        /* (mapstart+FOOGRID_IDX(map,e->Ycoord,e->Xcoord))->Tiletype=PTL_DOOR;
         NEED to be able to save and restore status of doors! */
        (mapstart+FOOGRID_IDX(map,e->Ycoord,e->Xcoord))->Tiletype=e->Open?PTL_ODOOR:PTL_DOOR; /* Sorted! */
    }
}

void build_tunnel(region_t *region, int xstart, int ystart, int xend, int yend)
{
    tunnellist_t *thissection;
    int x,y;
    maploc_t *mapstart=map->Data;
    tunnellist_t **tmapstart=tmap->Data; /*Newly added*/
    direction_t dir;
    thissection=(tunnellist_t *)counting_malloc(MEM_TUNNEL,sizeof(tunnellist_t));
    thissection->Xstart=xstart;thissection->Ystart=ystart;
    thissection->Xend=xend;thissection->Yend=yend;
    /* Ah. This upsets find_area(). which assumes start<=end. Hard to change*/
    if(xstart>xend){thissection->Xstart=xend; thissection->Xend=xstart;}
    if(ystart>yend){thissection->Ystart=yend; thissection->Yend=ystart;}
    thissection->Region=region;
    /* Now... we have to add this tunnel to... somewhere.
       Should be on a region. Now, regions should be created with the
       first tunnel of the region. OR, we could create a new region with
       each tunnel, and then call merge_regions() on both. Owie. I
       now have a headache.
       OK, Since decided that the caller specifies the region, which it
       decides whether or not to create. */
    thissection->Nexttunnel=region->Firsttunnel;region->Firsttunnel=thissection;

    if(xstart!=xend && ystart!=yend)
    {
      sprintf(strbuf,"X: %d Y: %d x: %d y: %d Not a straight line tunnel",xstart,ystart,xend,yend);
      bugger(REPTYPE_MAPGEN,LOGLEVEL_FOOBAR,strbuf);
      crash(OUT_OF_BOUNDS);
    }
    if((xstart==0&&xend==0)||(ystart==0&&yend==0))  {
      thissection=NULL;xstart=thissection->Xstart; /* Crash it here */
      crash(INCONSISTENT);
    }
      /* Means that the tunnel is not orthogonal*/
    if (xstart<xend) dir=FACE_E;
    else if (xstart>xend) dir=FACE_W;
    else if (ystart<yend) dir=FACE_S;
    else dir=FACE_N;
    switch(dir)
    {
        case FACE_N:
          x=xstart;
          for(y=ystart;y>=yend;y--)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=thissection;
          }
          break;
        case FACE_E:
          y=ystart;
          for(x=xstart;x<=xend;x++)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=thissection;
          }
          break;
        case FACE_S:
          x=xstart;
          for(y=ystart;y<=yend;y++)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=thissection;
          }
          break;
        case FACE_W:
          y=ystart;
          for(x=xstart;x>=xend;x--)
          {
             (mapstart+FOOGRID_IDX(map,y,x))->Tiletype=PTL_TUNNEL;
             *(tmapstart+FOOGRID_IDX(tmap,y,x))=thissection;
          }
          break;
        default:
          crash(FUNCTION_ERROR);
          break;
    }
}

void merge_areas(area_t *area1, area_t *area2)
{
   /* This *REQUIRES* that room[] is GLOBAL. The problem is, that
     means that unless we create it with malloc(), that array is
     going to just sit around the whole time whether we need it or not*/
   int rlist;
/* tunnellist_t *tun; */
   region_t *reg;
   area_t *next;
   area_t *scan;
   if (area1==area2) return; /*Nothing to do!*/
   if (NULL==area1)
   {
     show_messages("area1 is null!");
     crash(INCONSISTENT);
   }
   if (NULL==area2)
   {
     show_messages("area2 is null!");
     crash(INCONSISTENT);
   }
   for (rlist=area1->Firstroom;tls->Room[rlist].Nextroom!=-1;rlist=tls->Room[rlist].Nextroom)
      tls->Room[rlist].Area=area2;
   /*Now we have the end of list 1.*/
   tls->Room[rlist].Area=area2; /*I think we need that again? */
   tls->Room[rlist].Nextroom=area2->Firstroom;

   /*I think that mebbe area1 might not have any tunnel??
    Either, way, a segfault happens here!*/

   /* Right, now we have to alter this as the region code has
     tunnellists moved to region structures, but areas now have region lists. */
/* for (tun=area1->Firsttunnel;tun->Nexttunnel!=NULL;tun=tun->Nexttunnel) */
   if (area1->Firstregion!=NULL)
   {
     for (reg=area1->Firstregion;reg->Nextregion!=NULL;reg=reg->Nextregion)
     {
        /*
        tun->Area=area2; */
        reg->Area=area2;
     }
  /* tun->Area=area2; again, we repeat that  */
     reg->Area=area2;
  /* tun->Nexttunnel=area2->Firsttunnel; */
     reg->Nextregion=area2->Firstregion;
     area2->Firstregion=area1->Firstregion;
   } /* else { Would we DO anything else? } */
   area2->Firstroom=area1->Firstroom;
/* area2->Firsttunnel=area1->Firsttunnel; Now in merge_regions(). */
/* area2->Firstregion=area1->Firstregion;  Depends on area1 HAVING any. */
   area2->Size+=area1->Size;
   /* NOW we want to remove area1 from the area list. OUCH!
     THIS MUST BE DONE!*/
   next=area1->Nextarea; /* The bit we use to replace area1 in the list*/
   if (tls->Firstarea==area1)
   { /* The exception to the rule: area1 is head of the list.*/
      tls->Firstarea=next; /* That's all folks!*/
   } else {
     for(scan=tls->Firstarea;scan->Nextarea!=area1;scan=scan->Nextarea)
     {
        if(scan->Nextarea==NULL)
         {show_messages("Found end of area list!");}
     }
     /*We've now found scan->Nextarea==area1. So we splice here.*/
     scan->Nextarea=next;
   } /*area1 should now be out of the list. Yes?*/
   counting_free(MEM_AREA,area1); /*we *can* do that, right? */
   /*Hopefully there should be no pointers to area1 left now.*/
}

/* NEW FUNCTION for region modification. */
void merge_regions(region_t *region1, region_t *region2)
{
   /* Things to do:
   -Check region1 and region2 are different
   -Check neither region1 nor region2 is NULL
   -check region1->Area == region2->Area (you should call merge_areas first)
   -Go through all the tunnels in region1, and put them in region2
   -Go through all the entrances in region1, and put them in region2
      -NEW: Now do similar for intersections!
   -where merge_areas() used to remove area1 from the global area list,
    we, er, must remove region1 from region1->Area->Regionlist??
    That is OK, yes?
   -free region1.
   ...profit? */
   tunnellist_t *tun;
   entrancelist_t *ent;
   intersection_t *corn;
   area_t *mutualarea;
   region_t *next;
   region_t *scan;

   if (region1==region2) /* No excuse for this. */
   { show_messages("Merging region with itself!");
     crash(INCONSISTENT); /*It just shouldn't happen.*/
   }
   if (NULL==region1)
   { show_messages("region1 is null!");
     crash(INCONSISTENT);
   }
   if (NULL==region2)
   { show_messages("region2 is null!");
     crash(INCONSISTENT);
   }
   mutualarea=region1->Area;
   if (mutualarea!=region2->Area) /*Aargh! */
   { show_messages("tried to merge regions from different areas!");
     crash(INCONSISTENT);
   }
   /* Args are OK. Now merge tunnels. */
   /* Hey, does this routine go kaboom if either region has no tunnels?
      It prolly would, but by definition a region MUST have at least
      one tunnel and 2 entrances. */
   for (tun=region1->Firsttunnel;tun->Nexttunnel!=NULL;tun=tun->Nexttunnel)
   {
      tun->Region=region2;
   }
   tun->Region=region2; /* we repeat that */
   tun->Nexttunnel=region2->Firsttunnel;
   region2->Firsttunnel=region1->Firsttunnel;

   /* Tunnels merged. Now merge entrances. */
   for (ent=region1->Firstentrance;ent->Next!=NULL;ent=ent->Next)
   {
      ent->Entrance->Region=region2; /* I think that's right... */
   }
   ent->Entrance->Region=region2; /* Repeat it */
   ent->Next=region2->Firstentrance;
   region2->Firstentrance=region1->Firstentrance;
   /* There's no Size element for regions like for areas, is there? */

   corn=region1->Intersections;
   if(corn!=NULL) {
     for(corn=region1->Intersections;corn->Next!=NULL;corn=corn->Next);
     corn->Next=region2->Intersections;
     region2->Intersections=region1->Intersections;
     /* Is all that OK? */
   }

   /* Entrances merged. Now remove region1 from the area's regionlist. */
   next=region1->Nextregion; /* We replace region1 with this in the list */
   if (mutualarea->Firstregion==region1)
   { /* The exception to the rule. */
      mutualarea->Firstregion=next; /* And that's all! */
   } else { /* everything else */
      for (scan=mutualarea->Firstregion;scan->Nextregion!=region1;scan=scan->Nextregion)
      {
         if(scan->Nextregion==NULL)
         {show_messages("Found end of region list!");}
      }
      /* We've now found scan->Nextregion==region1, so we splice here. */
      scan->Nextregion=next;
   } /* So region1 should now not appear in the region list now. Right? */

   /* We should now be OK to free region1. */
   counting_free(MEM_REGION,region1);
}


/* CHANGED to now only get area of tunnels (due to a rare bug
  if get_area found a room in front of the path when a probe
  was actually stopped by a tunnel next to it */
/* Also, this currently wants to check through tunnellists to work
  out area of a tunnel- but this was the stupid, old way, that was
  before I put in the grid of tunnel pointers. If I just use that
  grid instead, the region change shouldn't hurt here. */
region_t *find_region(int x, int y)
{
/* area_t *testarea=firstarea;  No longer used? */
/* tunnellist_t *testtun;  No longer used? */
   maploc_t *mapstart=map->Data;
   tunnellist_t **tmapstart=tmap->Data; /* Is finally used now.*/
   tunnellist_t *gridtun; /* Used as a temp var */
   if ((mapstart+FOOGRID_IDX(map,y,x))->Tiletype==PTL_PWALL)
   {
      show_messages("Permanent wall found!!");
      crash(OUT_OF_BOUNDS);
      /* Should we instead return NULL? */
   }

   /* Should only respond to tunnel sections. But what about entrances,
     which *are* part of regions?? */
   if ((mapstart+FOOGRID_IDX(map,y,x))->Tiletype==PTL_TUNNEL||(mapstart+FOOGRID_IDX(map,y,x))->Tiletype==PTL_DOOR||(mapstart+FOOGRID_IDX(map,y,x))->Tiletype==PTL_ODOOR)
   {
     /* This is a tunnel.  We *USED TO* have to search all
      tunnel sections... but now we have the grid of tunnel-pointers.*/
     gridtun=*(tmapstart+FOOGRID_IDX(tmap,y,x));
     if (NULL!=gridtun) return gridtun->Region; /* Rrrggghh how strenuous... */

     /*
     for(;testarea!=NULL;testarea=testarea->Nextarea)
     {
        if(testarea==NULL) crash(INCONSISTENT);
        for(testtun=testarea->Firsttunnel;testtun!=NULL;testtun=testtun->Nexttunnel)
        {
          if (testtun==NULL) crash(INCONSISTENT);
          if(testtun->Xstart<=x && testtun->Xend>=x &&
            testtun->Ystart<=y && testtun->Yend>=y)
             return(testtun->Area);
        }
     }
     */

     /*It looks like this doesn't happen *any more*... */
     crash(INCONSISTENT); return(NULL); /*To shut up gcc*/
   } else {
    /* Technically it might not be, but we'll assume it's part of a room.*/
    /* This can now be given from room[find_room(x,y)]->Area;
       So we should remove this? YES!! Now just returns NULL
       if no tunnel found. */
     return(NULL);
   }
}

int find_room(int x,int y)
{
   int r;
   for (r=0;r<217;r++)
   {
       if(tls->Room[r].Xcentre!=-1 &&
       (tls->Room[r].Xstart<=x && tls->Room[r].Xstart+tls->Room[r].Width-1>=x &&
        tls->Room[r].Ystart<=y && tls->Room[r].Ystart+tls->Room[r].Height-1 >=y))
            return(r);
   }
   bugger(REPTYPE_CODE,LOGLEVEL_FOOBAR,"find_room() called for a place that wasn't a room. That should never be done.");
   return(-1); /*Not part of any room we can find!*/
}

entrance_t *find_door(region_t *region,int x,int y)
{
   /* search through region->Firstentrance, for entrance at x,y */
   tile_t tile;
   entrancelist_t *scan;
   entrance_t *door;
   for(scan=region->Firstentrance;scan!=NULL;scan=scan->Next) {
      door=scan->Entrance;
      if(door->Xcoord==x && door->Ycoord==y) return(door);
   }

   /* WHOA! It wasn't there. This shouldn't happen. */
   tile=(((maploc_t *)map->Data)+FOOGRID_IDX(map,y,x))->Tiletype;
   if (tile==PTL_DOOR||tile==PTL_ODOOR) {
      sprintf(strbuf,"find_door:Door found at %d,%d wasn't found in specified region.",x,y);
      bugger(REPTYPE_MAPGEN,LOGLEVEL_WARNING,strbuf);
      if (find_region(x,y)==region) {
         sprintf(strbuf,"...The region at %d,%d WAS the specified region. The map data's screwed.",x,y);
         bugger(REPTYPE_MAPGEN,LOGLEVEL_FOOBAR,strbuf);
         crash(FUNCTION_ERROR);
      }
      for(scan=find_region(x,y)->Firstentrance;scan!=NULL;scan=scan->Next) {
         door=scan->Entrance;
         if(door->Xcoord==x && door->Ycoord==y) return(door);
      }
      bugger(REPTYPE_MAPGEN,LOGLEVEL_FOOBAR,"...Still can't find it. This is screwed up.");
      crash(FUNCTION_ERROR);
   } else {
      sprintf(strbuf,"find_door:There was NO door found at %d,%d.",x,y);
      bugger(REPTYPE_MAPGEN,LOGLEVEL_FOOBAR,strbuf);
      crash(FUNCTION_ERROR);
   }
   return(NULL); /* To shut up GCC */
}

entrance_t *find_common_door(region_t *region,int roomno)
{
   /* We search through the region for the room, rather than
   searching the room for the region. */
   entrancelist_t *scan;
   if (NULL==region) return(NULL); /* Lazy evaluation doesn't seem to work! */
   for(scan=region->Firstentrance;scan!=NULL;scan=scan->Next)
       if (scan->Entrance->Roomno==roomno) return(scan->Entrance);
   return(NULL);
   /* Ahh. */
}

void replace_items(item_t *item)
{
   /* PROBLEM: We really want to put the highest item in a heap
      on top of the heap, as maploc->Itemlist.
      But that would require us to make numerous passes over the
      list, to first put the items whose Next element is NULL, and
      then pass to find items whose Next element is the item
      previously placed, etc etc.
      Alternatively, we could redo the Next elements, and leave
      the heaps out of order.
      But *would* they be out of order? Things dropped not only
      go to the top of the heap, but also top of the itemlist.
      So strictly, the heaps should be in the same order as the
      itemlist!
      So if we traverse the itemlists in order, and place items
      on the *bottom* of their heaps, the heaps will be in order.
      Appending to the bottom of lists is icky. Alternatives:
      place items on top of their heaps, reverse heap order after
      all placed. Quicker? Probably. Harder to code? YES!
      Could possibly speed up the problem of appending to ends
      of lists by having a small cache of coordinates that have
      large heaps, and having pointers to the ends of their lists */
   /* Endcache arrangement: ([x&3]+(4*[y&3]))*2;
      So, 16 distinct sets of coords, each set can store
      cache for 2 actual coords, giving 32 cache elements */
   struct i_endcache_t {
      int X,Y,Size;
      item_t *Listend;
   };
   int i,x,y,size;
   item_t *prev;
   struct i_endcache_t endcache[32];
   maploc_t *mapstart=map->Data;
   maploc_t *elem;
   for(i=0;i<32;i++)
   {
      endcache[i].Size=0; /* elements of size 0 must be invalid */
   }

   for(;item!=NULL;item=item->Next2)
   {
      x=item->X;y=item->Y;
      i=(x&3)+(4*(y&3)); /* Index to start of endcache */
      /* If (x,y) is in the cache, we get the "prev" value
      from there, and then increment the Size element after
      appending our item to the list.
      If not, we work out the "prev" value by just traversing
      the heap, and then append our item. We remember the (new!)
      size of the heap.
      If that size is above, say, 2 items, we check the (two)
      potential elements of the cache to see if one is either
      vacant or has a lower size. If so, we cache the new
      details there. Otherwise, we just carry on. */
      elem=(mapstart+FOOGRID_IDX(map,item->Y,item->X));
      /* Check the cache. */
      if ((endcache[i].Size>0 &&endcache[i].X==x&&endcache[i].Y==y)
      ||(++i,(endcache[i].Size>0) &&endcache[i].X==x&&endcache[i].Y==y))
      {  /* That means the place IS in the cache, now at element i */
          (endcache[i].Listend)->Next=item; /* Append!*/
          item->Next=NULL; /* Yes indeedy */
          endcache[i].Size++; /* Increase heap size for cache element */
      } else { /* Wasn't in the cache at all */
          prev=elem->Itemlist;
          if (prev==NULL) { /* empty heap exception */
             elem->Itemlist=item;
             item->Next=NULL; /* Only item there so far */
             /* Don't bother caching this. */
          } else { /* heap */
             for (prev=elem->Itemlist,size=0;prev->Next!=NULL;prev=prev->Next,size++);
             prev->Next=item;
             item->Next=NULL;
             /* Now, see if we could/should stick this in the cache */
             if(size>2&&( /* If the size is v low, is it worth it? */
             (size>endcache[i].Size) || (++i,(size>endcache[i].Size)) )) {
                /* OK, put one in [i] */
                endcache[i].Listend=item;
                endcache[i].Size=size;
                endcache[i].X=x; endcache[i].Y=y;
             }
          } /* End of dealing with heaps that
          weren't in the cache */
      } /* Should now have placed the item. */
   }
   /* And that's all the items done :) */
}

void add_room_items(int roomno)
{
   /* NB: Until I track down the glitch that pops up in the
     first map generated, I'm NOT going to do anything that would
     cause that map to change.
     That includes using ANY sort of random numbers in any new functions.
     So for now, this function will not be remotely random.
     TADA! As I fixed aforementioned glitch, this is now randomized!*/
   int x,y,count=0,pick=0;
   itemtype_t *picktype; /* Pick a type. */
   itemgroup_t group; /* From a group. */
   if (roomno==-1|| tls->Room[roomno].Xcentre==-1) return; /* Hmm */
   /* Was (foo*roomno)/217 */
   x=tls->Room[roomno].Xstart+1 + (( (tls->Room[roomno].Width-2) *roll_dX(d20set[0]))/20);
/* y=room[roomno].Ycentre+1; */
   y=tls->Room[roomno].Ystart+1 + (( (tls->Room[roomno].Height-2)*roll_dX(d12set[0]))/12);
   /* Now pick a group. Was roomno%IG_LASTGROUP. */
   /* At time of writing, LASTGROUP was ~23, and code will need
     to be changed if it reaches 36 or so. */
   group=(IG_LASTGROUP*((roll_dX(d6set[0]))+(6*roll_dX(d6set[1]))) )/36;
   /* Make it something else if we picked a corpse.
     Was done as roomno%3?food:stone, which meant usually food. */
   if (group==IG_CORPSE) group=(roll_dX(d4set[0]))?IG_FOOD:IG_STONE;
   /* corpses are only generated from monsters. We've got none yet.*/

   /* 
   for(picktype=first_it;picktype!=NULL;picktype=picktype->Next)
   {
      if (picktype->Type==group)  Potential match.
      {
         flag=1;
         count++;
         chosen=picktype;
         if(count>(roomno/8)) break;
      } else if (flag==1) break;
   }
   */
   /* Now pick a specific item type from that group.
     Was (it_nums[group]*roomno)/217. Now it's random, we
     pick a number in the range 0<143, supposing some group such
     as weapons might have *lots* of item types. */
   pick=(it_nums[group]*((roll_dX(d12set[0]))+(12*roll_dX(d12set[1])) ))/144;
   for(picktype=first_its[group];count<pick;count++,picktype=picktype->Next);
   if (picktype==NULL) crash(OUT_OF_BOUNDS);
   add_item(tls,x,y,picktype);
}

void replace_mobs(monster_t *mob)
{
   maploc_t *mapstart=map->Data;
   for (;mob!=NULL;mob=mob->Next)
   {
      (mapstart+FOOGRID_IDX(map,mob->Y,mob->X))->Mptr=mob;
      /* That all? */
   }
}

void add_room_mobs(int roomno)
{
   /* This function *should* now be entirely random- The position in the
     room should be random, and the mob chosen should be random. Yes? */
   int x,y,count=0,pick=0;
   mobtype_t *picktype;
   if (roomno==-1|| tls->Room[roomno].Xcentre==-1) return; /* Hmm */
   /* Was (foo*roomno)/217 */
   x=tls->Room[roomno].Xstart+1 + (( (tls->Room[roomno].Width-2) *roll_dX(d20set[0]))/20);
   y=tls->Room[roomno].Ystart+1 + (( (tls->Room[roomno].Height-2)*roll_dX(d12set[0]))/12);
/* pick=(mt_nums*roomno)/217; */
   /* 64 is reasonable range as using only letters means maximum 52
      different choices. We will probably end up having each character
      be a *group* of monsters. At that point, we should have a separate
      "group" and "pick" like with item placement. */
   pick=(mt_nums*(roll_dX(d8set[0])+(8*roll_dX(d8set[1])) ))/64;
   for(picktype=first_mt;count<pick;count++,picktype=picktype->Next);
   if(picktype==NULL) crash(FUNCTION_ERROR);
   /* Now we have to set up its initial cache after creating it*/
   aicache_trycacheroom(add_mob(tls,x,y,picktype),roomno);
   /* Yes, we discard the return value. We dont need it yet. */
}

void remake_features(levelstructs_t *level)
{
   int i;
   maploc_t *mapstart=map->Data;
   maploc_t *elem;
   for (i=0;i<level->Nfeatures;i++)
   {
      elem=mapstart+FOOGRID_IDX(map,level->Features[i].Ycoord,level->Features[i].Xcoord);
      elem->Tiletype=level->Features[i].Feature;
/*    elem->Extra=level->Features[i].Extra; That now stays in the feature */
      elem->Featureidx=(char)i;
   }
}

int make_feature(int roomno,tile_t tile)
{
   int x,y;
   maploc_t *mapstart=map->Data;
   maploc_t *elem;
   featuredata_t feat;
   if (roomno==-1|| tls->Room[roomno].Xcentre==-1) crash(INCONSISTENT);
   if (tls->Nfeatures>9) crash(INCONSISTENT); /* Too many */
   x=tls->Room[roomno].Xstart+1 + (( (tls->Room[roomno].Width-2) *roll_dX(d20set[0]))/20);
   y=tls->Room[roomno].Ystart+1 + (( (tls->Room[roomno].Height-2)*roll_dX(d12set[0]))/12);
   tls->Features[tls->Nfeatures].Xcoord=x;
   tls->Features[tls->Nfeatures].Ycoord=y;
   tls->Features[tls->Nfeatures].Feature=tile;
   elem=(mapstart+FOOGRID_IDX(map,y,x));
   /* CHANGE to make sure the space is currently PTL_FLOOR or PTL_RMID,
   and if not, pick another space */
   elem->Tiletype=tile;
   elem->Featureidx=(char)tls->Nfeatures; /* Store an index here */
   switch (tile) {
      case PTL_PWALL:
      case PTL_ROCK:
         ;
         break;
      case PTL_STAIRSUP:
      case PTL_STAIRSDOWN:
         /* Set dest to NULL. After creating a location pointer! */
         feat.Destination=(locid_t *)counting_malloc(MEM_LOCID,sizeof(locid_t));
         if (feat.Destination==NULL) crash(NO_MEM);
         feat.Destination->Map=NULL;
/*       elem->Extra=feat;  Not stored on the map now */
         break;
      case PTL_ALTAR:
         /* Set to PGOD_EB / PGOD_GRUBBLE */
         feat.Deity=(roll_dX(d4set[0])<2)?PGOD_EB:PGOD_GRUBBLE;
/*       elem->Extra=feat; not stored on the map now */
         break;
      default:
         show_messages("make_feature() called with bad/stupid feature");
         crash(INCONSISTENT);
         break;
   }
   tls->Features[tls->Nfeatures].Extra=feat;
   tls->Nfeatures++;
   return(tls->Nfeatures-1);
}

area_t *choose_small_area()
{
   int minsize=218;   /*Min size found*/
   area_t *carea; /* area with that size*/
   area_t *tarea; /*Area we are testing.*/
/* static dXrng_roller *d4=NULL;
   if (NULL==d4) d4=new_dX(4); Using global dice array now */
   for (tarea=tls->Firstarea;tarea!=NULL;tarea=tarea->Nextarea)
   {
      if(tarea->Size==1)
      {
         minsize=1;carea=tarea;
         if (!roll_dX(d4set[0]))break; /*  We might stop looking here*/
      }
      if(tarea->Size<minsize)
      {
         minsize=tarea->Size;
         carea=tarea;
      }
   }
   return(carea);
}

/* STUFF FOR FREEING */

void free_tunnels(tunnellist_t *tl)
{
    tunnellist_t *next;
    for(;tl!=NULL;next=tl->Nexttunnel,counting_free(MEM_TUNNEL,tl),tl=next);
}

/* NB: Now tunnellists are in regions, so I don't know how we do the freeing.
  SIMPLE: Now we have a free_regions() that calls free_tunnels, and 
  gets called by free_areas. */
void free_intersections(intersection_t *ilist)
{
    intersection_t *next;
    for(;ilist!=NULL;next=ilist->Next,counting_free(MEM_INTERSECTION,ilist),ilist=next);
}

void free_entrancelist(entrancelist_t *elist,int free_entrance)
{
    /* set free_entrance=1 when freeing list from region;
    set to 0 when freeing from room. */
    entrancelist_t *next;
    for(;elist!=NULL;next=elist->Next,counting_free(MEM_ENTLIST,elist),elist=next)
    {
       if (free_entrance) counting_free(MEM_ENTRANCE,elist->Entrance);
    }
}

void free_regions(region_t *region)
{
    region_t *next;
    for(;region!=NULL;next=region->Nextregion,counting_free(MEM_REGION,region),region=next)
    {
       free_tunnels(region->Firsttunnel);
       /* We should also free the entrances. And the intersections. */
       free_intersections(region->Intersections);
       free_entrancelist(region->Firstentrance,1);
         /*also call that on room's entrancelists, to remove the *lists*
         but not the entrances themselves. The entrance data is shared,
         so we only free that once! */
    }
}

void free_features() {
    int i;
    for(i=0;i<tls->Nfeatures;i++) {
       switch (tls->Features[i].Feature) {
          case PTL_STAIRSUP:
          case PTL_STAIRSDOWN:
            counting_free(MEM_LOCID,tls->Features[i].Extra.Destination);
          default:
            ;
       }
    }
}

void free_areas()
{
    /*There should only be one area, firstarea.*/
    area_t *area;
    area_t *next;
    for(area=tls->Firstarea;area!=NULL;next=area->Nextarea,counting_free(MEM_AREA,area),area=next)
    {
       /*
       free_tunnels(area->Firsttunnel);
       */
       free_regions(area->Firstregion);
    }
}
