/* ==== monster.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include <stdlib.h>
#include "porrog_defs.h"
#include "C-foogrid.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h"
#include "player_t.h"
#include "ai_t.h"

/* protos */
#include "ai_p.h"
#include "monster.h"
#include "los.h"
#include "ui.h"
#include "item_p.h" /* For the corpses */
#include "mg_utils.h"
#include "memutils.h"
#include "combat.h"
#include "crash.h"

/* foogrid *map SHOULD be declared global in maparray.h anyway */
monster_t *add_mob(levelstructs_t *level,int x,int y,mobtype_t *mtype)
{
   monster_t *newmob;
   aicache_memory_t *mmem;
   long offset=FOOGRID_IDX(map,y,x);
   maploc_t *mapstart=map->Data;
   newmob=(monster_t *)counting_malloc(MEM_MONSTER,sizeof(monster_t));
   if (newmob==NULL) crash(NO_MEM);
   newmob->Mobtype=mtype;
   newmob->Inventlist=NULL; /*Carrying nothing.*/
   newmob->X=x;newmob->Y=y; /*Monster knows where it is*/
   newmob->Var_stats.MaxHp=10; /* Uh... */
   newmob->Var_stats.Hp = newmob->Var_stats.MaxHp;

   /* AI stuff */
   mmem=(aicache_memory_t *)counting_malloc(MEM_MMEM,sizeof(aicache_memory_t));
   if (mmem==NULL) crash(NO_MEM);
   newmob->Memory=mmem;
   mmem->Roomcache=NULL;
   mmem->Tunnelcache=NULL;
   /* Now buffer entries: Don't need setting up, just creating. */
   mmem->Roombuffer=(aicache_rooment_t *)counting_malloc(MEM_ROOMMEM,sizeof(aicache_rooment_t));
   if (mmem->Roombuffer==NULL) crash(NO_MEM);
   mmem->Roombuffer->Paths=NULL;
   mmem->Roombuffer->Next=NULL; /* To simplify freeing */
   mmem->Roombuf_flag=0; /* Invalid buffer */
   mmem->Tunnelbuffer=(aicache_tunnelent_t *)counting_malloc(MEM_TUNMEM,sizeof(aicache_tunnelent_t));
   if (mmem->Tunnelbuffer==NULL) crash(NO_MEM);
   mmem->Tunnelbuffer->Paths=NULL;
   mmem->Tunnelbuffer->Next=NULL;
   mmem->Tunnelbuf_flag=0;
   /* Other AI stuff */
   newmob->Ttype=1; /* Lets the AI acquire new target */
   newmob->Annoyance=0; /* Feeling easygoing! */
   newmob->Next=level->Moblist;
   level->Moblist=newmob;
   (mapstart+offset)->Mptr=newmob; /*Map knows where it is*/
   level->Nmobs++; /* Increase the count */
   return(newmob);
}

void new_mobtype(mobtype_t *space, char sym, char *name,char *desc)
{
    int l;
    char *cname;
    space->Symbol=sym;
    space->Name=name;
    space->Desc=desc;
    l=1+sprintf(strbuf,"%s corpse",name);/* Ah. */
    cname=(char *)counting_malloc(MEM_STRING,l*sizeof(char));
    sprintf(cname,"%s corpse",name);
    space->Corpsetype=push_itype(IG_CORPSE,cname); /* does allocation anyways*/

}

void push_mtype(char sym,char *name)
{
    mobtype_t *newmt;
    newmt=(mobtype_t *)counting_malloc(MEM_MTYPE,sizeof(mobtype_t));
    newmt->Next=first_mt; /* Should have an array of mobtypes
       one for each symbol, like push_itype does */
    first_mt=newmt;
    mt_nums++;
    new_mobtype(newmt,sym,name,"A monster with no description");
}

/* Was (foogrid *map,dx,dy,mob) */
void move_mob(int dx, int dy, monster_t *mob)
{
    monster_t *opp;
    maploc_t *mapstart=map->Data;
    int x,y,tx,ty;
    long offset;
    tile_t tile;
    x=mob->X;y=mob->Y;
    tx=x+dx;ty=y+dy;
    offset=FOOGRID_IDX(map,ty,tx);
    tile=(mapstart+offset)->Tiletype;
    /* Commented whilst debugging map!!!  */
    if(tile==PTL_ROCK||tile==PTL_VWALL||tile==PTL_HWALL||tile==PTL_PWALL||tile==PTL_DOOR||tile==PTL_CORNER)
    {
       if (in_los(player.X,player.Y,x,y)) {
         sprintf(strbuf,"The %s walks into the %s!",mob->Mobtype->Name,tile==PTL_DOOR?"door":"wall");
         show_messages(strbuf);
       }
       return;
    } /**/
    opp=(mapstart+offset)->Mptr;
    if(opp!=NULL) { /* Unusual... */
#ifdef CALM
       sprintf(strbuf,"Monster at %d,%d attacks something, despite CALM option!",x,y);
       bugger(REPTYPE_AI|REPTYPE_COMBAT,LOGLEVEL_WARNING,strbuf);
#endif
/*     if (in_los(player.X,player.Y,x,y)) {   THIS is handled in m_hit_m().
         sprintf(strbuf,"The %s hits the %s!",mob->Mobtype->Name,opp->Mobtype->Name);
         show_messages(strbuf);
       } */
       m_hit_m(mob,opp);
       return; /* We won't have monsters fight each other this way... Will we?
        Maybe we will change later if we can make sure they normally walk
        around each other if they don't mean to fight */
    }
    if (player.X==tx&&player.Y==ty) {
       m_hit_p(mob); /* Fight!! */
       return; /* No, we don't keep walking... */
    }
    (mapstart+FOOGRID_IDX(map,y,x))->Mptr=NULL;
    /*x+=dx;y+=dy; */
    mob->X=tx;mob->Y=ty;
    (mapstart+FOOGRID_IDX(map,ty,tx))->Mptr=mob;
}

/* Was (foogrid *map,monster_t *mob) and opened any doors.
 Now opens one specific door. */
void m_open_door(monster_t *mob, int dx,int dy)
{
    int x=mob->X,y=mob->Y;
    int done=0;
    entrance_t *door;
    maploc_t *mapstart=map->Data;
    if((mapstart+FOOGRID_IDX(map,y+dy,x+dx))->Tiletype==PTL_DOOR)
    {
       (mapstart+FOOGRID_IDX(map,y+dy,x+dx))->Tiletype=PTL_ODOOR;
       door=find_door(find_region(x+dx,y+dy),x+dx,y+dy);
       door->Open=1;/* So it's persistent */
       done=1;
    }
    if (done==1)
    {
      /* Only show anything if the door is in los.
        If the monster is in los, show "the mob opens the door".
        If it isn't, show "something opens the door". */
      if (in_los(player.X,player.Y,x+dx,y+dy)) {
        if (in_los(player.X,player.Y,x,y)) {
          sprintf(strbuf,"The %s opens a door",mob->Mobtype->Name);
        } else {
          sprintf(strbuf,"Something opens a door");
        }
        show_messages(strbuf);
      }
    } else {
      sprintf(strbuf,"Monster at %d,%d tries to open a door at %d,%d where there isn't one!!",x,y,x+dx,y+dy); /* shouldnt happen */
      bugger(REPTYPE_AI|REPTYPE_MONSTERS|REPTYPE_MAPGEN,LOGLEVEL_WARNING,strbuf);
    }
}

void m_get_item(monster_t *mob,item_t *item)
{
    /* Assumes item is individual: item->Next MUST be NULL*/
    if (item==NULL) return; /*Play it safe*/
    item->Next=mob->Inventlist;
    mob->Inventlist=item;
}
item_t *m_discard_item(monster_t *mob)
{
    /* Removes top item from monster's inventory. Returns it
      as an individual item.*/
    item_t *item=mob->Inventlist;
    if (item==NULL) return(NULL); /*Plausible*/
    mob->Inventlist=item->Next;
    item->Next=NULL;
    return(item);
}

/* Should already have map declared. */
void m_die(levelstructs_t *level, monster_t *victim)
{
    monster_t *next;
    monster_t *scan;
    aicache_memory_t *mmem=victim->Memory;
    maploc_t *mapstart=map->Data;
    int x=victim->X, y=victim->Y;
    add_item(level,x,y,victim->Mobtype->Corpsetype);
    (mapstart+FOOGRID_IDX(map,y,x))->Mptr=NULL; /* KABOOM! */
    /* We should free this, yes? But if in a
    list of all monsters, need to free it from there too */
    /* Yup. Doing that now. */
    next=victim->Next;
    if(level->Moblist==victim) {
       level->Moblist=next; /* The exception */
    } else {
       for (scan=level->Moblist;scan->Next!=victim;scan=scan->Next);
       scan->Next=next;
    } /* Should now have removed victim from list. */
    level->Nmobs--;
    /* Now remove AI cache contents!!! */
    aicache_freerooments(mmem->Roombuffer);
    aicache_freetunnelents(mmem->Tunnelbuffer);
    aicache_freerooments(mmem->Roomcache);
    aicache_freetunnelents(mmem->Tunnelcache);
    counting_free(MEM_MMEM,mmem);
    counting_free(MEM_MONSTER,victim);
}

int m_whatamion(monster_t *mob)
{
    maploc_t *mapstart=map->Data;
    tile_t floor;
    floor=(mapstart+FOOGRID_IDX(map,mob->Y,mob->X))->Tiletype;
    switch (floor) {
       case PTL_PWALL:
       case PTL_ROCK:
       case PTL_HWALL:
       case PTL_VWALL:
       case PTL_CORNER:
          bugger(REPTYPE_MONSTERS|REPTYPE_AI,LOGLEVEL_FOOBAR,"Monster in a wall??");
          crash(INCONSISTENT);
          break; /*Not really needed*/
       case PTL_TUNNEL:
          return(0);
          break; /*Not really needed*/
       case PTL_DOOR:
       case PTL_ODOOR:
          return(2);
          break;
       case PTL_FLOOR:
       default: /* Includes features, etc */
          return(1);
          break;
    }
    return(3); /* To shut up gcc... */
}

int m_respect(monster_t *observer,monster_t *observed)
{
    /* How patient is observer with observed? */
    return(20);
    /* When the proper game mechanics are implemented, it'll
    work out something like (10*observed->Level)/oberserver->Level
    or something very similar. */
}
