/* ==== ui.c ==== */
/*  Part of the source of the program "Porrog", a roguelike game.
    Copyright (C) 2002-2003  Tom Barnes-Lawrence
    Current web address: www.angelfire.com/super2/duologue/
    Current email address: tomble@usermail.com
    Postal address: mind your own business

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


/*#include <stdlib.h>
*/

#include <string.h> /* Was included from ui.h, but strcpy etc
   are mostly just used here. */
#include <ncurses.h>
#include "porrog_defs.h"
#include "C-foogrid.h"
#include "memutils.h"
#include "item_t.h"
#include "structures.h"
#include "mapgen_types.h"
#include "maparray.h" /*We need this, but maybe just for the #defines*/
#include "player_t.h"

/* protos */
#include "item_p.h"
#include "crash.h"
#include "dXrng.h"
#include "los.h"
#include "player_p.h"
#include "apply.h"
#include "ui.h"


/* char strbuf[160]; Now in globals.c */
#define discardwin(win) werase(win);wrefresh(win);delwin(win);touchwin(stdscr);refresh()


void draw_status()
{
   char linebuf[29];
   int len;

   sprintf(linebuf,"X,Y:%d,%d",player.X,player.Y);
   len=strlen(linebuf);
   strncat(linebuf+len," ||::--::||  --::||::--  ||::",29-len);
   move(0,0);
   addstr(linebuf);

   sprintf(linebuf,"HP:%d/%d",player.Var_stats.Hp,player.Var_stats.MaxHp);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(3,0);
   addstr(linebuf);


   /* Base stats */
   sprintf(linebuf,"STR: %d",player.Fixed_stats.Str);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(8,0);
   addstr(linebuf);
   sprintf(linebuf,"DEX: %d",player.Fixed_stats.Dex);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(9,0);
   addstr(linebuf);
   sprintf(linebuf,"AGI: %d",player.Fixed_stats.Agi);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(10,0);
   addstr(linebuf);
   sprintf(linebuf,"CON: %d",player.Fixed_stats.Con);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(11,0);
   addstr(linebuf);

   /* Derived stats */
   sprintf(linebuf,"Ev: %d",player.Var_stats.Ev);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(13,0);
   addstr(linebuf);
   sprintf(linebuf,"Coverage: %d%%",player.Var_stats.Cov);
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(14,0);
   addstr(linebuf);
   sprintf(linebuf,"Prot Avg: %d",(player.Var_stats.Pravg/100));
   len=strlen(linebuf);
   strncat(linebuf+len,"                             ",29-len);
   move(15,0);
   addstr(linebuf);
}

void draw_map(int cornerx,int cornery,int px,int py)
{
   /* We assume that curses is already initialised. */
   int x,y,sx,sy,tile;
   maploc_t *elem;
   maploc_t *mapstart=map->Data;
   char sym;
#ifndef AUTOPILOT
   calc_los(map,px,py);
#else
   calc_los(map,tls->Moblist->X,tls->Moblist->Y);
#endif
   /* used to be ... ,sy=1; sy<23; ... */
   for (y=cornery,sy=0;sy<24;y++,sy++)
      for (x=cornerx,elem=mapstart+FOOGRID_IDX(map,y,x),sx=30;sx<80;elem++,sx++,x++)
      {
          move(sy,sx);
#ifndef AUTOPILOT
          if (in_los(px,py,x,y))
#else
          if (in_los(tls->Moblist->X,tls->Moblist->Y,x,y))
#endif
          {
           elem->Seen=1;
           elem->Remembered=NULL; /* Update it whenever square seen */
           if(NULL!=elem->Mptr) sym=elem->Mptr->Mobtype->Symbol;
           else if (px==x&&py==y) sym=player.Symbol;
           else if (NULL!=elem->Itemlist)
           {
              sym=elem->Itemlist->Itemtype->Symbol;
              elem->Remembered=elem->Itemlist->Itemtype;
           } else{
             tile=elem->Tiletype; /* show the floor */
             switch(tile)
             {
                 case PTL_PWALL:
                 case PTL_CORNER:
                   sym='#';break;
                   sym='*';break;
                 case PTL_ROCK:
                   sym='#';break;
                 case PTL_HWALL:
                   sym='#';break;
                   sym='-';break;
                 case PTL_VWALL:
                   sym='#';break;
                   sym='|';break;
                 case PTL_DOOR:
                   sym='+';break;
                 case PTL_ODOOR:
                   sym=':';break; /* Er, I suppose*/
                 case PTL_TUNNEL:
                   sym=':';break;
                 case PTL_STAIRSUP:
                   sym='<';break;
                 case PTL_STAIRSDOWN:
                   sym='>';break;
                 case PTL_ALTAR: /* And maybe some others? */
                   sym='_';break;
                 case PTL_FLOOR:
                 case PTL_RMID:
                   sym='.';break;
                 default:
                   sym=''; /* Foobar!!!*/
             }
           }
          } else
          { /* draw from memory */
            if(elem->Seen)
            {
               if(NULL!=elem->Remembered) sym=elem->Remembered->Symbol;
               else switch (elem->Tiletype) {
                       case PTL_PWALL:
                       case PTL_CORNER:
                          sym='#';break;
                          sym='*';break;
                       case PTL_ROCK:
                          sym='#';break;
                       case PTL_HWALL:
                          sym='#';break;
                          sym='-';break;
                       case PTL_VWALL:
                          sym='#';break;
                          sym='|';break;
                       case PTL_DOOR:
                          sym='+';break;
                       case PTL_ODOOR:
                          sym=':';break;
                       case PTL_TUNNEL:
                          sym=':';break;
                       case PTL_STAIRSUP:
                          sym='<';break;
                       case PTL_STAIRSDOWN:
                          sym='>';break;
                       case PTL_ALTAR:
                          sym='_';break;
                       case PTL_FLOOR:
                       case PTL_RMID:
                          sym='.';break;
                       default:
                          sym='';
               }
            } else sym=' '; /* blank! */
          }
          addch(sym);
      }
   refresh();
}

void show_messages(char *msg)
{
#define MESSAGESPACE_TOP 18
   /* Message area starts from line MESSAGESPACE_TOP, and runs from column 0
     to column 29. I think. Or is it 28? 
     Top was originally set to 18, may be changed to give more space. */
   char linebuf[29];
   char *startptr=msg;
   char *safeptr; /* When we find a bad line, this will be the end.*/
   char *testptr;
   int msglen; /*=strlen(msg); */
   int line;
   int done=0;
   static int startline=MESSAGESPACE_TOP,startcol=0,refr=0;
   if (msg==NULL) {
     startline=MESSAGESPACE_TOP;startcol=0;refr=1;return;/* Reset the thing */
   }
   /* Put this message in the ringbuffer */
   if (messages_ringbuffer[mrb_endidx]!=NULL) counting_free(MEM_STRING,messages_ringbuffer[mrb_endidx]);
   messages_ringbuffer[mrb_endidx]=counting_malloc(MEM_STRING,sizeof(char)*(1+strlen(msg)));
   strcpy(messages_ringbuffer[mrb_endidx],msg);
   mrb_endidx++; if(mrb_endidx==200) mrb_endidx=0;
   if (mrb_endidx==mrb_startidx) { /* Wrapped */
      mrb_startidx++; if (mrb_startidx==200) mrb_startidx=0;
   }
   do {
      for (line=startline;line<22;line++,startcol=0)
      {
         msglen=strlen(startptr);
         safeptr=startptr+msglen-1; /* why did I do this? I forget.*/
         testptr=strchr(startptr,' ');
         if ((testptr==NULL)&&msglen>27) crash(IO_ERROR);/*see below*/
         if ((testptr-startptr)>27) crash(IO_ERROR); /*Cant do owt wi that*/
         if ((testptr==NULL)&&msglen>(27-startcol)) continue;
         if ((testptr-startptr)>(27-startcol)) continue;/*Start on next line*/
         for(;
           testptr!=NULL&&(testptr-startptr)<(28-startcol);
           safeptr=testptr+1,testptr=strchr(safeptr,' '))
         ;

    /*if (testptr==NULL) msglen=strlen(startptr);  */
         if (startcol+msglen>29) msglen=(safeptr-startptr)-1;
         else done=1;
/* This needs much working out. Except it does seem to work now */
         if(startcol+msglen>29) crash(FUNCTION_ERROR); /*Arrgh!*/
         move(line,startcol);
         strncpy(linebuf,startptr,msglen);
         *(linebuf+msglen)='\0'; /*Add a null.*/
         strncat(linebuf,"                               ",29-(msglen+startcol));
         addstr(linebuf);
         refresh(); /* Should be done at the END of the routine.*/
         /* Was (testptr==NULL) */
         if(done) {startcol+=msglen+2;break;}
         startptr=safeptr;
      }
      startline=line;
      if(refr) {/* Clear the rest */
        for (line=startline+1;line<22;line++)
        {
           move(line,0);
           addstr("                              ");
        }
        refr=0;
      }
      if (startline>=22&&(!done)) { /* Hit the end. */
         move(23,1); /* in certain terminals you can use 24, but not all */
         addstr("[SPACE for more]");
         while(getch()!=' ');
         move(23,1);
         addstr("                ");
         startline=MESSAGESPACE_TOP;startcol=0;refr=1;
      }
   } while(/* testptr!=NULL*/ !done) ;
}

/* Read the messages ringbuffer (usu with ^P) */
void show_mrb()
{
   /* Not sure how to do the scrolling... */
   WINDOW *win;
   int i,rbi,tot,bline,line;
   int k;
   win=newwin(22,76,1,2);

   for(rbi=mrb_startidx,tot=0;rbi!=mrb_endidx;rbi++,tot++)
      if (rbi==200) rbi=0;
   /* Now have  tot == number of lines in ringbuffer */
   bline=(tot-19);if(bline<0) bline=0;
   line=bline;
   /* draw frame */
   for(i=1;i<21;i++) {
     wmove(win,i,0);waddch(win,'|');
     wmove(win,i,75);waddch(win,'|');
   }
   for(i=1;i<75;i++) {
     wmove(win,0,i);waddch(win,'-');
     wmove(win,21,i);waddch(win,'-');
   }
   wmove(win,0,30);waddstr(win,"Message history:");
   wmove(win,0,0);waddch(win,'+');wmove(win,0,75);waddch(win,'+');
   wmove(win,21,0);waddch(win,'+');wmove(win,21,75);waddch(win,'+');
   /* main display loop */
   do {
      /* werase(win); Or blank each line? */
      for(i=1,rbi=mrb_startidx+line;i<20&&rbi!=mrb_endidx;i++,rbi++) {
         if (rbi>=200) rbi-=200;
         wmove(win,i,1);
         if (messages_ringbuffer[rbi]==NULL) break; /* End of thingy */
         if(strlen(messages_ringbuffer[rbi])>73) sprintf(messages_ringbuffer[rbi]+70,"...");
         waddstr(win,messages_ringbuffer[rbi]);wclrtoeol(win);
         wmove(win,i,75);waddch(win,'|');
      }
      wmove(win,20,21);
      waddstr(win,"k:up j:down <SPACE>:return to game");
      touchwin(win);wrefresh(win);
      k=getch();
      if('k'==k && line>0) line--;
      if('j'==k && line<bline) line++;
   }
   while(' '!=k); /* Space to exit */
   /* Now remove the window */
   werase(win);wrefresh(win); delwin(win); /* clear+remove this window */

   touchwin(stdscr); refresh(); /* Refresh main window */
}

void free_mrb()
{
   int rbi;
   for(rbi=mrb_startidx;rbi!=mrb_endidx; rbi++) {
      if(rbi==200) rbi=0;
      if (messages_ringbuffer[rbi]==NULL) {
         fprintf(stderr,"free_mrb() found a NULL pointer.\n");
         break;
      }
      counting_free(MEM_STRING,messages_ringbuffer[rbi]);
   }
   if(messages_ringbuffer[rbi]!=NULL) counting_free(MEM_STRING,messages_ringbuffer[rbi]);
   /* Check this is good */
}

void msg_window(char *msg)
{
   int x,y;
   int length=strlen(msg);
   int xe=43+(length/2);
   for(x=37-(length/2);x<xe;x++)
   {
      move(8,x);addch('-');
      move(12,x);addch('-');
   }
   for(y=9;y<12;y++)
   {
      move(y,37-(length/2));addch('|');
      move(y,xe);addch('|');
   }
   move(8,37-(length/2));addch('+');
   move(12,37-(length/2));addch('+');
   move(8,xe);addch('+');
   move(12,xe);addch('+');
   move(10,40-(length/2));addstr(msg);
   move(14,0);
   refresh();
}

item_t *prompt_for_inv_item(int flag)
{ /* Tricky one. Takes list, displays it, asks you to pick one.*/
  /* Problem: Lists of items in your inventory are handled
     differently to items on the floor or in containers.
     Solution: give 2 functions:
         prompt_for_inv_item()
     and prompt_for_heap_item(item_t *list). */
/* int line;   Don't we use this?? */
   char k;
   item_t *item;
   if (NULL==player.Inventlist) {
     show_messages("You're not carrying anything!");
     return(NULL);
   }
   show_messages(flag?"Wield what? ? to choose item from list, - to wield nothing, ESC to cancel (FIXME)":"Which item? ? to choose from list.");
   while(1){
     k=getch();
     if (k=='?') { /* Show the list first */
        /*
        for(line=1,item=player.Inventlist;item!=NULL;line++,item=item->Next)
        {
           if(line<22) {move(line,0);
             addch(item->Invletter); addch(':');
             addstr(item->Itemtype->Name);}
        }
        refresh(); */
        display_heap("Items available",player.Inventlist); /* Tidier now */
     } else { /* Got a choice! If valid, return it. */
        if (k=='-'&&flag) return(phonyitem);
        if ((k>='a'&&k<='z')||(k>='A'&&k<='Z'))
        {
           for (item=player.Inventlist;item!=NULL;item=item->Next)
              if(item->Invletter==k) return(item);
           show_messages("You don't have *that*!");
        }
        if (k==27) /* ESC */ return(NULL);
     }
   }
}

item_t *prompt_for_heap_item(item_t *list)
{
   int line,items,choice,i;
   char k=' ';
   item_t *item;
   WINDOW *win;
   show_messages("Choose an item");
   refresh();
   while(list!=NULL) /* Er... */
   {
     for(items=0,item=list; items<15&&item!=NULL; items++,item=item->Next);
     win=newwin(items+4,60,9-(items/2),10);
     for(line=0,item=list; line<items &&item!=NULL; line++, item=item->Next)
     {
        wmove(win,line+2,0);waddch(win,'|');
        waddch(win,line+'a'); waddch(win,':');
        waddch(win,item->Itemtype->Symbol);waddch(win,' ');
        waddstr(win,item->Itemtype->Name);/*waddstr(win,"      ");  Should be better.*/
        /* Newly copied from display_heap() */
        wmove(win,line+2,59);waddch(win,'|');
     }
     wmove(win,0,0);waddch(win,'+');
     wmove(win,0,59);waddch(win,'+');
     wmove(win,line+3,0);waddch(win,'+');
     wmove(win,line+3,59);waddch(win,'+');
     wmove(win,1,0);waddch(win,'|');
     wmove(win,1,59);waddch(win,'|');
     wmove(win,line+2,0);waddch(win,'|');
     wmove(win,line+2,59);waddch(win,'|');
     wmove(win,0,1);for(i=0;i<58;i++)waddch(win,'-');
     wmove(win,line+3,1);for(i=0;i<58;i++)waddch(win,'-');
/*   for(;line<15;line++)  Blank the rest 
     { move(line+1,3);addstr("                        ");  That'll do. 
     } */
     touchwin(win);wrefresh(win); /* Display the window...*/
     while(1)
     {
       k=getch();
       if ('a'<=k && ('a'+items)>k) /* Was source of foobar... */
       {  /* Got one. Which was it? */
          choice=(k-'a');
          for(line=0,item=list; line<choice&&item!=NULL;line++,item=item->Next);
          if(NULL==item) crash(FUNCTION_ERROR); /* Shouldn't happen */
          discardwin(win);
          return(item);
       }
       if(27==k) {discardwin(win);return (NULL);} /* Don't choose anything */
       if(' '==k) break; /* Next page. */
     }
     /* User must have pressed <space> to choose next page. */
     list=item;
     discardwin(win);
   }
   /* GCC complains about control reaching here. Should it be
      returning NULL here? Or not reaching this point?? I'll look
      at it later. I suspect we should return NULL anyway. */
   return (NULL);
}

void display_heap(char *title,item_t *list)
{
   int count=0,x=0;
   item_t *scan;
   WINDOW *win;
   for(scan=list;scan!=NULL;scan=scan->Next)
      count++;
   if(count==0)return;
   /*
   move(8,1);
   addch('0'+(count/10));
   addch('0'+(count%10));
   refresh(); */
   win=newwin(count+4,60,9-(count/2),10);
   wmove(win,1,30-(strlen(title)/2));waddstr(win,title);
   for(count=0,scan=list;scan!=NULL;scan=scan->Next,count++)
   {
     wmove(win,count+2,0);waddch(win,'|');
     waddch(win,scan->Invletter);waddch(win,':');
     waddch(win,scan->Itemtype->Symbol);waddch(win,' ');
     waddstr(win,scan->Itemtype->Name);
     wmove(win,count+2,59);waddch(win,'|');
   }
   wmove(win,0,0);waddch(win,'+');
   wmove(win,0,59);waddch(win,'+');
   wmove(win,count+3,0);waddch(win,'+');
   wmove(win,count+3,59);waddch(win,'+');
   wmove(win,1,0);waddch(win,'|');
   wmove(win,1,59);waddch(win,'|');
   wmove(win,count+2,0);waddch(win,'|');
   wmove(win,count+2,59);waddch(win,'|');
   for(x=1;x<59;x++)
   {  wmove(win,0,x);waddch(win,'-');
      wmove(win,count+3,x);waddch(win,'-');
   }
   wmove(win,count+2,30-(strlen("Press <space> to continue")/2));
   waddstr(win,"Press <space> to continue");
   touchwin(win);
   wrefresh(win);
   while(' '!=getch())
   {}
   werase(win);wrefresh(win); /* blank it!*/
   delwin(win);

   touchwin(stdscr); /* I have to do this to restore what was under the */
   refresh(); /* window I just removed. Is this really the *only* way?? */

}

/* This will prolly never be used, as the odd characters don't seem to
be remotely portable. */
char arcanify(char n)
{
   if(n>='A'&&n<='Z') n-=('A'-'a'); /* to lowercase */
   if(n>='a'&&n<='z')
   {
      return(*("ߢef9hjk|Mn5+vWץZ"+(n-'a')));
   }
   return(n);
}

void apply_container(item_t *container)
{
   item_t *item;
   item_t *contents=container->Instdata.Container.Contents;
   char k; /* key */
   int openable=container->Itemtype->Typedata.Container.Opens;
   char *cname=container->Itemtype->Name;
   if (openable) {
      sprintf(strbuf,"You open the %s,",cname);
      show_messages(strbuf);
   }
   /* if container's empty, ask if player wants to put something in.
   Else ask if they want to take something out, put something in, or
   do nothing. */
   if (contents==NULL) { /* empty */
      sprintf(strbuf,"The %s is empty. Do you want to put something in?",cname);
      show_messages(strbuf);
      k=getch();
      if(k!='y'&&k!='Y') {
        show_messages("OK, so don't.");
        return;
      }
      item=prompt_for_inv_item(0);
      if(NULL==item) { show_messages("Cancelled.");return;}
      if(item==container) { show_messages("That's STUPID!");return;}
      if (can_go_in_container(container,item)) {
        p_discard_item(item);
        put_in_container(container,item);
        sprintf(strbuf,"You put the %s in the %s.",item->Itemtype->Name,cname);
        show_messages(strbuf);
      }
   } else { /* Here we get a choice! */
      sprintf(strbuf,"There's something in the %s!",cname);
      show_messages(strbuf);
      sprintf(strbuf,"t: take something out. p: put something in. anything else: cancel");
      show_messages(strbuf);
      k=getch();
      switch(k) {
         case 't':
         case 'T':
           item=prompt_for_heap_item(contents);
           if(NULL==item) { show_messages("So don't then.");return;}
           remove_from_container(container,item);
           p_get_item(item); /* Not sure this is right. Meh. */
           sprintf(strbuf,"You take the %s out of the %s.",item->Itemtype->Name,cname);
           show_messages(strbuf);
           break;
         case 'p':
         case 'P':
           item=prompt_for_inv_item(0);
           if(NULL==item) { show_messages("Cancelled.");return;}
           if(item==container) { show_messages("That's STUPID!");return;}
           if (can_go_in_container(container,item)) {
             p_discard_item(item);
             put_in_container(container,item);
             sprintf(strbuf,"You put the %s in the %s.",item->Itemtype->Name,cname);
             show_messages(strbuf);
           }
           break;
         default:
           show_messages("OK.");
      }
   }
}
